/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  ekf.c
  8 December 1992 (originated)

  AUTHOR:
    Ali Azarbayejani

  CONTENTS:
    ekf_print()
    ekf_matlab_dump()
    ekf_size_in_bytes()
    ekf_malloc()
    ekf_fill_given_memory()
    ekf_make_filter()
    ekf_clear_state()
    ekf_resize_filter()
    ekf_filter_step()

  DESCRIPTION:
    Based on:
    kalman.c -- Routines for the Extended Kalman Filter
	  		Author:	Martin R. Friedmann
	  		Date:	Mon Mar 25 12:41:08 EST 1991
    Modified Marty-ware to allow
      (1) variable state and signal vector sizes,
      (2) hooks to functions for replacing default state
          estimation and prediction,
      (3) fast scalar EKFs

  REFERENCES:
    Brown, Robert G., << Introduction to Random Signal Analysis and Kalman
        Filtering >>, John Wiley & Sons, New York, 1983

    Gelb, Arthur (Editor), << Applied Optimal Estimation >>, MIT Press,
       Cambridge, MA, 1974

    Friedmann, Martin, Alex Pentland, and Thad Starner, "Device
       Synchronization using an Optimal Linear Filter", Proceedings of the
       ACM Symposium on Interactive 3D Graphics, Cambridge, MA, Vol 26, No
       2, March 1992

    Azarbayejani, Ali and Alex Pentland, "Recursive Estimation of Motion,
       Structure, and Focal Length", IEEE PAMI, June 1995

*/

#include <math.h>
#include <stdio.h>
#include "matrix/matrix.h"
#include "ekf.h"

/* have to change these to sub_matrix_print_blurb()
 * as soon as they're written
 */
void
  ekf_print(FILE *stream, ekf_type *f)
{
  fprintf(stderr, "Vector %s at %x\n", "Y", f->Y);
  fprintf(stderr, "Vector %s at %x\n", "Y_p", f->Y_p);
  fprintf(stderr, "Vector %s at %x\n", "dY", f->dY);
  fprintf(stderr, "Vector %s at %x\n", "X_p", f->X_p);
  fprintf(stderr, "Vector %s at %x\n", "dX", f->dX);
  fprintf(stderr, "Vector %s at %x\n", "X_e", f->X_e);

  matrix_print_blurb( stream, f->I->M );
  matrix_print_blurb( stream, f->P_p->M );
  matrix_print_blurb( stream, f->P_e->M );
  matrix_print_blurb( stream, f->H->M );
  matrix_print_blurb( stream, f->Q->M );
  matrix_print_blurb( stream, f->R->M );
  matrix_print_blurb( stream, f->K->M );
  matrix_print_blurb( stream, f->Phi->M );
  matrix_print_blurb( stream, f->TPxP->M );
  matrix_print_blurb( stream, f->TPxN->M );
  matrix_print_blurb( stream, f->TNxN->M );
  matrix_print_blurb( stream, f->TNxP->M );
}

void
  ekf_matlab_dump(FILE *str, ekf_type *f, char *label)
{
  char	symbol[32];

  sprintf(symbol, "xp%s", label);
  print_matlab_vector(str, f->X_p, f->n, symbol);
  sprintf(symbol, "y%s", label);
  print_matlab_vector(str, f->Y, f->p, symbol);
  sprintf(symbol, "Pp%s", label);
  print_matlab_sub_matrix(str, f->P_p, symbol);
  sprintf(symbol, "H%s", label);
  print_matlab_sub_matrix(str, f->H, symbol);
  sprintf(symbol, "K%s", label);
  print_matlab_sub_matrix(str, f->K, symbol);
  sprintf(symbol, "Q%s", label);
  print_matlab_sub_matrix(str, f->Q, symbol);
  sprintf(symbol, "R%s", label);
  print_matlab_sub_matrix(str, f->R, symbol);
}
 
int
  ekf_size_in_bytes(int n_state, int n_sig)
{
  int 	NxN 	= flex_matrix_size_in_bytes( n_state, n_state );
  int 	PxP 	= flex_matrix_size_in_bytes( n_sig, n_sig );
  int 	PxN 	= flex_matrix_size_in_bytes( n_sig, n_state );
  int 	NxP 	= flex_matrix_size_in_bytes( n_state, n_sig );
  int 	size 	= sizeof(ekf_type);
  
  size += 3 * sizeof(FLOATTYPE) * n_sig;	/* three signal vectors */
  size += 3 * sizeof(FLOATTYPE) * n_state;  	/* three state vectors */
  
  size += NxN;		/* I:	Identity Matrix */
  size += NxN; 		/* P_p: Error Covariance Pred. */
  size += NxN;		/* P_e: Error Covariance Est.  */
  size += PxN; 		/* H: 	Linearized Signal */
  size += NxN;		/* Q: 	Dynamics Noise Covariance */
  size += PxP; 		/* R: 	Signal Noise Covariance*/
  size += NxP; 		/* K: 	Kalman Gain Matrix */
  size += NxN; 		/* Phi: Linearized State */
  size += PxP; 		/* Temp Matrix */
  size += PxN; 		/* Temp Matrix */
  size += NxP; 		/* Temp Matrix */
  size += NxN; 		/* Temp Matrix */

  return size;
}

static ekf_type *
  ekf_malloc(int n_state, int n_sig)
{
  return((ekf_type *) calloc (1, ekf_size_in_bytes(n_state, n_sig)));
}

void
  ekf_fill_given_memory(ekf_type *f,
			int 	n,
			int 	p,
			void	(*initialize)(),
			void	(*pre_process)(),
			void	(*update_H)(),
			void	(*predict_signal)(),
			void	(*estimate_state)(),
			void	(*update_Phi)(),
			void	(*predict_state)(),
			void	(*post_process)(),
			char	*name)
{
  int 	NxN 	= flex_matrix_size_in_bytes( n, n );
  int 	PxP 	= flex_matrix_size_in_bytes( p, p );
  int 	PxN 	= flex_matrix_size_in_bytes( p, n );
  int 	NxP 	= flex_matrix_size_in_bytes( n, p );
  char 	*cp;

  cp = (char *)(f + 1);

  f->Y = (FLOATTYPE *) cp;		cp += sizeof(FLOATTYPE) * p;
  f->Y_p = (FLOATTYPE *) cp;		cp += sizeof(FLOATTYPE) * p;
  f->dY = (FLOATTYPE *) cp;		cp += sizeof(FLOATTYPE) * p;
  f->X_p = (FLOATTYPE *) cp;		cp += sizeof(FLOATTYPE) * n;
  f->dX = (FLOATTYPE *) cp;		cp += sizeof(FLOATTYPE) * n;
  f->X_e = (FLOATTYPE *) cp;		cp += sizeof(FLOATTYPE) * n;

  f->I = (flex_matrix_type *) cp;	cp += NxN;
  f->P_p = (flex_matrix_type *) cp;	cp += NxN;
  f->P_e = (flex_matrix_type *) cp; 	cp += NxN;
  f->H = (flex_matrix_type *) cp;	cp += NxP;
  f->Q = (flex_matrix_type *) cp;	cp += NxN;
  f->R = (flex_matrix_type *) cp;	cp += PxP;
  f->K = (flex_matrix_type *) cp;	cp += PxN;
  f->Phi = (flex_matrix_type *) cp;	cp += NxN;
  f->TPxP = (flex_matrix_type *) cp;	cp += PxP;
  f->TPxN = (flex_matrix_type *) cp;	cp += PxN;
  f->TNxP = (flex_matrix_type *) cp;	cp += NxP;
  f->TNxN = (flex_matrix_type *) cp;	cp += NxN;

  f->name = name;
  f->n = f->N_max = n;
  f->p = f->P_max = p;

#define FMTGM fill_flex_matrix_type_given_memory
    FMTGM(f->I,		n, n, "Identity");
    FMTGM(f->P_p, 	n, n, "Error Covariance Prediction");
    FMTGM(f->P_e, 	n, n, "Error Covariance Estimate");
    FMTGM(f->H, 	n, p, "Linearized Measurement Model");
    FMTGM(f->Q, 	n, n, "Dynamics Error Covariance");
    FMTGM(f->R, 	p, p, "Measurement Error Covariance");
    FMTGM(f->K, 	p, n, "Kalman Gain Matrix");
    FMTGM(f->Phi, 	n, n, "STM Matrix");
    FMTGM(f->TPxP,	p, p, "Temp Matrix PxP");
    FMTGM(f->TPxN, 	n, p, "Temp Matrix PxN");
    FMTGM(f->TNxN, 	n, n, "Temp Matrix NxN");
    FMTGM(f->TNxP, 	p, n, "Temp Matrix NxP" );
#undef FMTGM

    f->initialize = initialize;
    f->pre_process = pre_process;
    f->linearize_signal_model = update_H;
    f->predict_signal= predict_signal;
    f->estimate_state = estimate_state;
    f->make_state_transition = update_Phi;
    f->predict_state = predict_state;
    f->post_process = post_process;

    matrix_identity( f->I->M );
}


ekf_type *
  ekf_make_filter(int	n_sig_max,	/* maximum number signals       */
		  int	n_st_max,	/* maximum number states	*/
		  void	(*initialize)(),/* Init xp, Pp, Q, R, H, Phi	*/
		  void	(*pre_process)(), /* Can prepare xp,Pp,Q,R,Phi  */
		  void	(*update_H)(),  /* Opt: if NULL, have to init H */
		  void	(*update_Phi)(),/* Opt: if NULL, have to init Phi */
		  void	(*predict_signal)(),	/* Optional	*/
		  void	(*estimate_state)(),	/* Optional	*/
		  void	(*predict_state)(),	/* Optional	*/
		  void	(*post_process)(),	/* Optional	*/
		  char	*name)
{
    register ekf_type *f;
    
    if (n_sig_max <= 0 || n_st_max <= 0 || initialize == NULL) {
	fprintf(stderr, "ekf_init_filter: Error in arguments!");
	return NULL;
    }

    f = ekf_malloc(n_st_max, n_sig_max);

    ekf_fill_given_memory(f,
			  n_st_max,
			  n_sig_max,
			  initialize,
			  pre_process,
			  update_H,
			  predict_signal,
			  estimate_state,
			  update_Phi,
			  predict_state,
			  post_process,
			  name);

    return f;
}

void
  ekf_clear_state(register ekf_type *f)
{
    vector_zero( f->Y, f->p );
    vector_zero( f->Y_p, f->p );
    vector_zero( f->dY, f->p );
    vector_zero( f->X_p, f->n );
    vector_zero( f->dX, f->n );
    vector_zero( f->X_e, f->n );

    matrix_zero( f->P_p->M );
    matrix_zero( f->P_e->M );
    matrix_zero( f->H->M );
    matrix_zero( f->Q->M );
    matrix_zero( f->R->M );
    matrix_zero( f->K->M );
    matrix_zero( f->Phi->M );
    matrix_zero( f->TPxP->M );
    matrix_zero( f->TPxN->M );
    matrix_zero( f->TNxP->M );
    matrix_zero( f->TNxN->M );

    matrix_identity( f->I->M );
}

void
  ekf_resize_filter(ekf_type *f, int n, int p)
{
  if(n == f->n && p == f->p)
    return; /* NO ERROR */

  if(n < 0 || n > f->N_max || p < 0 || p > f->P_max)
    return; /* ERROR */
  
  f->n = n;
  f->p = p;
  sub_matrix_reframe(f->I, 0, 0, n, n);
  sub_matrix_reframe(f->P_p, 0, 0, n, n);
  sub_matrix_reframe(f->P_e, 0, 0, n, n);
  sub_matrix_reframe(f->H, 0, 0, p, n);
  sub_matrix_reframe(f->Q, 0, 0, n, n);
  sub_matrix_reframe(f->R, 0, 0, p, p);
  sub_matrix_reframe(f->K, 0, 0, n, p);
  sub_matrix_reframe(f->Phi, 0, 0, n, n);
  sub_matrix_reframe(f->TPxP, 0, 0, p, p);
  sub_matrix_reframe(f->TPxN, 0, 0, p, n);
  sub_matrix_reframe(f->TNxP, 0, 0, n, p);
  sub_matrix_reframe(f->TNxN, 0, 0, n, n);

  f->TPxP->line = f->TPxP->M->w = p;	/* to allow for efficient inversion */
  f->TPxP->M->h = p;
}

void
  ekf_filter_step(ekf_type *f, void *params)
{
  if(!f)
    return;

  if(f->pre_process)
    (*f->pre_process)(f, params);
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
   * At this point, we have x(t|t-1), P(t|t-1), and y(t)       *
   *   and Q(t), R(t).					       *
   * Also must have					       *
   *    H(t) or  f->linearize_signal_model != NULL	       *
   *    Phi(t) or f->state_transition != NULL		       *
  \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

  if(f->linearize_signal_model)				/* f->H(t) */
    (*f->linearize_signal_model)(f, params);

  sub_matrix_multiply( f->H, f->P_p, f->TPxN );		/* f->K(t) */
  sub_matrix_multiply_transpose( f->TPxN, f->H, f->TPxP );
  sub_matrix_add( f->TPxP, f->R, f->TPxP );
  sub_matrix_invert( f->TPxP, f->TPxP );
  sub_matrix_transpose_multiply( f->H, f->TPxP, f->TNxP );
  sub_matrix_multiply( f->P_p, f->TNxP, f->K );

  if(f->predict_signal)					/* f->Y_p(t) */
    (*f->predict_signal)(f, params);
  else 
    sub_matrix_multiply_vector( f->H, f->X_p, f->Y_p);

  if(f->estimate_state)					/* f->X_e(t) */
    (*f->estimate_state)(f, params);
  else {
    vector_subtract( f->Y, f->Y_p, f->p, f->dY );
    sub_matrix_multiply_vector( f->K, f->dY, f->dX );
    vector_add( f->X_p, f->dX, f->n, f->X_e );
  }

  sub_matrix_multiply( f->K, f->H, f->TNxN );		/* f->P_e(t) */
  sub_matrix_subtract( f->I, f->TNxN, f->TNxN );
  sub_matrix_multiply( f->TNxN, f->P_p, f->P_e );

  if(f->make_state_transition)				/* f->Phi(t) */
    (*f->make_state_transition)(f, params);
  if(f->predict_state)					/* f->X_p(t+1) */
    (*f->predict_state)(f, params);
  else
    sub_matrix_multiply_vector( f->Phi, f->X_e, f->X_p );

  sub_matrix_multiply( f->Phi, f->P_e, f->TNxN );	/* f->P_p(t+1) */
  sub_matrix_multiply_transpose( f->TNxN, f->Phi, f->P_p );
  sub_matrix_add( f->P_p, f->Q, f->P_p );

  if(f->post_process)
    (*f->post_process)(f, params);
}
