/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  matrix.c
  Mon Mar 11 03:49:39 EST 1991

  AUTHOR:
    Martin Friedmann

  CONTENTS:
    Min()
    Max()
    matrix_multiply_error()
    matrix_multiply()
    matrix_transpose_multiply()
    matrix_multiply_transpose()
    matrix_multiply_vector_n()
    matrix_multiply_vector()
    matrix_add()
    matrix_subtract()
    matrix_scale()
    vector_scale()
    vector_add()
    vector_subtract()
    vector_zero()
    vector_fill()
    vector_copy()
    matrix_invert()
    matrix_invert_lu_too()
    matrix_cmp()

  DESCRIPTION:
    This file contains routines for adding, multiplying and printing
    matrices and vectors.

*/

#include "matrix.h"

#define Min(x,y) (((x) < (y)) ? (x) : (y))
#define Max(x,y) (((x) > (y)) ? (x) : (y))

void matrix_multiply_error( m1, m2, m_out, routine )
    matrix_type *m1, *m2, *m_out;
    char * routine;
{
    fprintf(stderr, "%dx%d matrix %s X %dx%d matrix %s = %dx%d matrix %s\n",
	    m1->h, m1->w, m1->name,
	    m2->h, m2->w, m2->name,
	    m_out->h, m_out->w, m_out->name);
    fprintf(stderr, "are incompatible for %s\n", routine);
}

/****************************************************************************
 *  matrix_multiply:  performs m_out = m1 X m2; 
 ****************************************************************************/
void matrix_multiply(m1, m2, m_out)
    matrix_type *m1, *m2, *m_out;
{
    register int k, width1, width2;
    register FLOATTYPE sum, *rowptr1, *rowptr2, *ptr_out;
    register int j, i;

    if ( m1->w != m2->h || m2->w != m_out->w || m1->h != m_out->h ) {
	matrix_multiply_error( m1, m2, m_out, "matrix_multiply" );
	return;
    }

    width1 = m1->w;
    width2 = m2->w;
    ptr_out = m_out->m;
    
    for (i = 0; i < m1->h; i++)          /* step down  */
	for (j = 0; j < width2; j++)         /* step across  */
	{
	    sum = 0.0;
	    rowptr1 = (FLOATTYPE *) (m1->m + i * width1);
	    rowptr2 = (FLOATTYPE *) (m2->m + j);
	    k = width1;
	    Duff8(k,
		  sum += *rowptr1++ * *rowptr2;
		  rowptr2 += width2);
	    *ptr_out++ = sum;
	}
}

/****************************************************************************
 *  matrix_transpose_multiply:  performs m_out = transpose(m1) X m2; 
 ****************************************************************************/
void matrix_transpose_multiply(m1, m2, m_out)
    matrix_type *m1, *m2, *m_out;
{
    register int i, j, k, width1, width2, height1;
    register FLOATTYPE sum, *rowptr1, *rowptr2, *ptr_out;
    
    if (m1->h != m2->h || m2->w != m_out->w || m1->w != m_out->h ) {
	matrix_multiply_error( m1, m2, m_out, "matrix_transpose_multiply" );
	return;
    }
    
    height1 = m1->h;
    width1 = m1->w;
    width2 = m2->w;
    ptr_out = m_out->m;
    
    for (i = 0; i < width1; i++) {
	for (k = 0; k < width2; k++) {
	    sum = 0.0;
	    rowptr1 = m1->m + i;
	    rowptr2 = m2->m + k;
	    j = height1;
	    Duff8( j,
		  sum += *rowptr1 * *rowptr2;
		  rowptr1 += width1;
		  rowptr2 += width2 );
	    *ptr_out++ = sum;
	}
    }
}

/****************************************************************************
 *  matrix_multiply_transpose:  performs m_out = m1 X transpose(m2); 
 ****************************************************************************/
void matrix_multiply_transpose(m1, m2, m_out)
    matrix_type *m1, *m2, *m_out;
{
    register int i, j, k, width1, height1, height2;
    register FLOATTYPE sum, *rowptr1, *rowptr2, *ptr_out;
    
    if (m1->w != m2->w || m2->h != m_out->w || m1->h != m_out->h ) {
	matrix_multiply_error( m1, m2, m_out, "matrix_multiply_transpose" );
	return;
    }
    
    width1 = m1->w;
    height1 = m1->h;
    height2 = m2->h;
    ptr_out = m_out->m;

    for (i = 0; i < height1; i++) {
	for (k = 0; k < height2; k++) {
	    rowptr1 = (FLOATTYPE *) (m1->m + ( i * width1 ));
	    rowptr2 = (FLOATTYPE *) (m2->m + ( k * width1 ));
	    sum = 0.0;
	    j = width1;
	    Duff8( j, sum += *rowptr1++ * *rowptr2++ );
	    *ptr_out++ = sum;
	}
    }
}

/****************************************************************************
 *  matrix_multiply_vector_n:  performs v_out = m1 v ; 
 ****************************************************************************/
void matrix_multiply_vector_n(m, v, n, v_out)
    int n;
    matrix_type *m;
    FLOATTYPE *v, *v_out;
{
    register int k, width;
    register FLOATTYPE sum, *rowptr, *vptr;
    register int i;

    width = m->w;
    
    if ( m->h < n ) {
	fprintf( stderr, "matrix_multiply_vector_n Error!\n");
	return;
    }
    rowptr = (FLOATTYPE *) m->m;
    for (i = 0; i < n; i++)          /* step down  */
    {
	sum = 0.0;
	vptr = v;
	k = width;
	Duff8( k, sum += *rowptr++ * *vptr++ );
	*v_out++ = sum;
    }
}
/****************************************************************************
 *  matrix_multiply_vector:  performs v_out = m1 v ; 
 ****************************************************************************/
void matrix_multiply_vector(m, v, v_out)
    matrix_type *m;
    FLOATTYPE *v, *v_out;
{
    register int k, width;
    register FLOATTYPE sum, *rowptr, *vptr;
    register int i;

    width = m->w;
    
    rowptr = (FLOATTYPE *) m->m;
    for (i = 0; i < m->h; i++)          /* step down  */
    {
	sum = 0.0;
	vptr = v;
	k = width;
	Duff8( k, sum += *rowptr++ * *vptr++ );
	*v_out++ = sum;
    }
}

/****************************************************************************
 * matrix_add: Adds two matrices together...  
 ****************************************************************************/
void matrix_add(m1, m2, m_out)
matrix_type *m1, *m2, *m_out;
{
    register FLOATTYPE *p1, *p2, *out;
    register int size = m1->w * m1->h;

#ifdef DEBUG
    if (m1->w != m2->w || m1->w != m_out->w ||
	m1->h != m2->h || m1->h != m_out->h ) {
	fprintf(stderr, "error: %s = %s + %s\n",
		m_out->name, m1->name, m2->name);
	return;
    }
#endif

    p1 = m1->m;
    p2 = m2->m;
    out = m_out->m;
    Duff8( size, *out++ = *p1++ + *p2++ );
}
/****************************************************************************
 * matrix_subtract: subtracts two matrices together...  OK if (m? == m_out) 
 ****************************************************************************/
void matrix_subtract(m1, m2, m_out)
matrix_type *m1, *m2, *m_out;
{
    register int size = m1->w * m1->h;
    register FLOATTYPE *mm1 = m1->m, *mm2 = m2->m, *out = m_out->m;

#ifdef DEBUG
    if (m1->w != m2->w || m1->w != m_out->w ||
	m1->h != m2->h || m1->h != m_out->h ) {
	fprintf(stderr, "error: %s = %s - %s\n",
		m_out->name, m1->name, m2->name);
	return;
    }
#endif

    Duff8( size, *out++ = *mm1++ - *mm2++ );
}
/****************************************************************************
 * matrix_scale: Scales matrix by constant  OK if (m == m_out) 
 ****************************************************************************/
void matrix_scale(m, c, m_out)
FLOATTYPE c;
matrix_type *m, *m_out;
{
    register int size = m->w * m->h;
    register FLOATTYPE *mm1 = m->m, *out = m_out->m;
    
    if (m->w != m->w || m_out->h != m_out->h ) {
	fprintf(stderr, "error: %s = %s * %g\n",
		m_out->name, m->name, c);
	return;
    }
    Duff8( size, *out++ = *mm1++ * c );
}

void vector_scale(v, c, len, out)
FLOATTYPE *v, c, *out;
int len;
{
    Duff8( len, *out++ = *v++ * c );
}

void vector_add(v1, v2, len, out)
FLOATTYPE *v1, *v2, *out;
int len;
{
    Duff8( len, *out++ = *v1++ + *v2++);
}

void vector_subtract(v1, v2, len, out)
FLOATTYPE *v1, *v2, *out;
int len;
{
    Duff8( len, *out++ = *v1++ - *v2++);
}

void vector_zero(v1, len)
FLOATTYPE *v1;
int len;
{
    Duff8( len, *v1++ = 0.0);
}

void vector_fill(v1, len, value)
FLOATTYPE *v1, value;
int len;
{
    Duff8( len, *v1++ = value);
}

void vector_copy(v1, v2, len)
FLOATTYPE *v1, *v2;
int len;
{
    Duff8( len, *v2++ = *v1++);
}


/****************************************************************************
 * matrix_invert: Performs matrix inversion using gaussian elimination.
 *		Because lu_decomp is used, the matrix m is copied for the
 * 		computation.  If you wish to obtain the LU decomposition
 * 		also, call matrix_invert_lu_too or write it if I havent.
 */
void matrix_invert( m, minverse )
matrix_type *m, *minverse;
{
    int i, j, k;
    register FLOATTYPE *p, *pe, *xp, *b, *y, *x;
    matrix_type *copy_of_m;
    FLOATTYPE bvec[90];
    
    if (m->w != m->h || m->w != minverse->w || m->w != minverse->h) {
	fprintf( stderr, "matrices %s cannot be inverted to %s\n", m->name,
		minverse->name);
	return;
    }

    /* special case this, it's just too good to be true */
    if (matrix_bandwidth(m) == 0) {
	if (matrix_bandwidth(minverse) != 0) {
	    matrix_zero(minverse);
	    matrix_bandwidth(minverse) = 0;
	}
	pe = minverse->m + m->w * Min(m->w, m->h);
	for (xp = m->m, p = minverse->m; p < pe; xp += m->w,p += m->w) {
	    *p++ = ((*xp == 0.0) ? 9999999999999.0 : 1.0 / *xp);
	    xp++;
	}
	return;
    }
    pe = minverse->m + m->w * m->h;
    
    copy_of_m = copy_matrix_type( m );
    lu_decomp(copy_of_m);

    if (m->w > 30) {
	b = (FLOATTYPE *) malloc (sizeof(FLOATTYPE) * m->w * 3);
	if (b == NULL) {
	    fprintf(stderr, "malloc failed: matrix_invert %dx3\n", m->w);
	    return;
	}
    }
    else
	b = bvec;
 
    y = &b[m->w];
    x = &y[m->w];
    
    for (i = 0; i < m->w; i++) 
	b[i] = 0.0;

    for (i = 0; i < m->w; i++) {
	b[i] = 1.0;
	forward_substitute(copy_of_m, b, y);
	backward_substitute(copy_of_m, y, x);
	b[i] = 0.0;

	xp = x;
	for (p = minverse->m + i; p < pe; p += m->w)
	    *p = *xp++;
    }
    if (m->w > 30) 
	free(b);
    free_matrix_type( copy_of_m );
}
/****************************************************************************
 * matrix_invert_lu_too: Performs matrix inversion using gaussian elimination.
 * 		Its just like matrix_invert, but it destroys (decomposes) the
 * 		input matrix m.
 */
void matrix_invert_lu_too( m, minverse )
matrix_type *m, *minverse;
{
    int i, j, k;
    register FLOATTYPE *p, *pe, *xp, *b, *y, *x;
    FLOATTYPE bvec[90];
    
    if (m->w != m->h || m->w != minverse->w || m->w != minverse->h) {
	fprintf( stderr, "matrices %s cannot be inverted to %s\n", m->name,
		minverse->name);
	return;
    }

    pe = minverse->m + m->w * m->h;
    
    lu_decomp(m);
    if (m->w > 30) 
	b = (FLOATTYPE *) malloc (sizeof(FLOATTYPE) * m->w * 3);
    else
	b = bvec;
    
    y = &b[m->w];
    x = &y[m->w];
    
    for (i = 0; i < m->w; i++) 
	b[i] = 0.0;

    for (i = 0; i < m->w; i++) {
	b[i] = 1.0;
	forward_substitute(m, b, y);
	backward_substitute(m, y, x);
	b[i] = 0.0;

	xp = x;
	for (p = minverse->m + i; p < pe; p += m->w)
	    *p = *xp++;
    }
    if (m->w > 30) 
	free(b);
}

/****************************************************************************
 * matrix_cmp: Scream if two matrices are different...  
 ****************************************************************************/
int matrix_cmp(m1, m2)
matrix_type *m1, *m2;
{
    register FLOATTYPE *p1, *p2;
    register int size = m1->w * m1->h;
    int same = 1;

#ifdef DEBUG
    if (m1->w != m2->w || m1->w != m_out->w ||
	m1->h != m2->h || m1->h != m_out->h ) {
	fprintf(stderr, "error: %s == %s\n",
		m_out->name, m1->name, m2->name);
	return;
    }
#endif

    p1 = m1->m;
    p2 = m2->m;
    Duff8( size, if (*p1++ != *p2++ ) same = 0 );
    if (!same)
	fprintf(stderr, "%s != %s\n", m1->name, m2->name );
    return same;
}
