/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  matrix.h
  Mon Mar 11 03:49:39 EST 1991

  AUTHOR:
    Martin Friedmann

  CONTENTS:
    fabs()
    matrix_type
    matrix_width()
    matrix_height()
    matrix_name()
    matrix_bandwidth()
    matrix_data()
    matrix_row_order()
    MATRIX_ARRAY_DEF()
    MREF_QUICK()
    MREF()
    AREF()
    vector_dot3()
    vector_dot4()
    vector_cross3()
    vector_scale3()
    vector_add3()
    vector_subtract3()
    Duff4()
    Duff8()

  DESCRIPTION:
    This file contains routines for adding, multiplying and printing
    matrices and vectors as well as routines for solving linear systems.
    The routine lu_decomp performs dumb gaussian elimination on a matrix A
    storing both resulting matrices L and U in the space originally
    occupied by A.  The routine forward_substitute then solves Ly = b, and
    the routine backward_substitute solves the final Ux = y.  So we solve
    the system Ax = b in the following way:
    
    WHAT	HOW
    
    A = LU	lu_decomp(A, width, height)	  \* A is now L & U *\
    Ly = b	forward_substitute(A, w, h, b, y) \* Solve for Y *\
    Ux = y	backward_substitute(A, w, h, y, x)\* Solve for X *\

*/

#ifndef _MATRIX_H_
#define _MATRIX_H_

#include <math.h>
#include <stdio.h>

#ifndef fabs
#define fabs(x) (((x) < 0.0) ? -(x) : (x))
#endif

#define FLOATTYPE float
#define SMALLFLOAT 1.0E-14
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

typedef struct {
    int w;		/* matrix width */
    int h;		/* matrix height */
    char *name;		/* name of matrix */
    int bandwidth;	/* max upper or lower diagonal with non-zero */
    FLOATTYPE *m;	/* m[height][width] */
    int free_data;	/* flag telling whether to free m[][] */
    int *row_order;	/* array with row-reorderings */
} matrix_type;

#define matrix_width(matrix) (matrix)->w
#define matrix_height(matrix) (matrix)->h
#define matrix_name(matrix) (matrix)->name
#define matrix_bandwidth(matrix) (matrix)->bandwidth
#define matrix_data(matrix) (matrix)->m
#define matrix_row_order(matrix) (matrix)->row_order

/* To use MATRIX_ARRAY_DEF( w, sym );
 * sym = (FLOATYPE **) matrix_data(mat);
 *
 * sym [i][j];
 */
#define MATRIX_ARRAY_DEF(w, sym) FLOATTYPE *(sym[w]);

/****************************************************************************
 * MREF and MREF_QUICK:  Reference Matrix...  MREF_QUICK assumes the existance
 * 			of a variable named "w" which contains the width of
 * 			the matrix.
 ****************************************************************************/
#define MREF_QUICK(mat, y, x) (*((mat)->m + (y) * w + (x)))
#define MREF(mat, y, x) (*((mat)->m + (y) * (mat)->w + (x)))

#define AREF(array, y, x, w) (*((array)+(y)*(w)+(x)))


#define vector_dot3(v1, v2)\
    (FLOATTYPE) ( (v1)[0] * (v2)[0] + (v1)[1] * (v2)[1] + (v1)[2] * (v2)[2] )

#define vector_dot4(v1, v2)\
    (FLOATTYPE) ( (v1)[0] * (v2)[0] + (v1)[1] * (v2)[1] + (v1)[2] * (v2)[2] + (v1)[3] * (v2)[3] )

#define vector_cross3(v1, v2, vout) {\
    (vout)[0] = ((v1)[1] * (v2)[2]) - ((v1)[2] * (v2)[1]);\
    (vout)[1] = ((v1)[2] * (v2)[0]) - ((v1)[0] * (v2)[2]);\
    (vout)[2] = ((v1)[0] * (v2)[1]) - ((v1)[1] * (v2)[0]);}

#define vector_scale3(v, s, vout) {\
    (vout)[0] = (v)[0] * (s);\
    (vout)[1] = (v)[1] * (s);\
    (vout)[2] = (v)[2] * (s);}

#define vector_add3(v1, v2, vout) {\
    (vout)[0] = (v1)[0] + (v2)[0];\
    (vout)[1] = (v1)[1] + (v2)[1];\
    (vout)[2] = (v1)[2] + (v2)[2];}

#define vector_subtract3(v1, v2, vout) {\
    (vout)[0] = (v1)[0] - (v2)[0];\
    (vout)[1] = (v1)[1] - (v2)[1];\
    (vout)[2] = (v1)[2] - (v2)[2];}

extern matrix_type *make_matrix_type();
extern matrix_type *copy_matrix_type();
extern matrix_type *make_matrix_type_w_data();
extern void fill_matrix_type_w_data();
extern void fill_matrix_type_given_memory();
extern void free_matrix_type();
extern void matrix_identity();
extern void matrix_zero();

extern void matrix_multiply();
extern void matrix_transpose_multiply();
extern void matrix_multiply_transpose();
extern void matrix_transpose();
extern void matrix_add();
extern int matrix_cmp();
extern void matrix_subtract();
extern void matrix_scale();
extern void vector_add();
extern void vector_subtract();
extern void vector_scale();

extern void vector_print();
extern void matrix_print();
extern void vector_print_non_zeros();
extern void matrix_print_non_zeros();
extern void vector_print_bool();
extern void matrix_print_bool();

extern int lu_decomp();
extern void forward_substitute();
extern void backward_substitute();
extern void matrix_invert();
extern void matrix_invert_lu_too();


/* A generalized form of a x4 Duff's Device */
#define Duff4(counter, block) { \
  while (counter >= 4) {\
     { block; } \
     { block; } \
     { block; } \
     { block; } \
     counter -= 4; \
  } \
     switch (counter & 3) { \
     case 3:	{ block; } \
     case 2:	{ block; } \
     case 1:	{ block; } \
     case 0: \
     counter = 0; \
   } \
}

/* A generalized form of a x8 Duff's Device */
#define Duff8(counter, block) { \
  while (counter >= 8) {\
     { block; } \
     { block; } \
     { block; } \
     { block; } \
     { block; } \
     { block; } \
     { block; } \
     { block; } \
     counter -= 8; \
  } \
     switch (counter & 7) { \
     case 7:	{ block; } \
     case 6:	{ block; } \
     case 5:	{ block; } \
     case 4:	{ block; } \
     case 3:	{ block; } \
     case 2:	{ block; } \
     case 1:	{ block; } \
     case 0: \
     counter = 0; \
   } \
}

#endif
