/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  matrix_init.c
  Mon Mar 11 03:49:39 EST 1991

  AUTHOR:
    Martin Friedmann

  CONTENTS:
    Min()
    Max()
    matrix_init_row_order()
    matrix_size_in_bytes()
    matrix_malloc()
    _fill_matrix_type()
    copy_matrix_type()
    make_matrix_type()
    make_matrix_type_w_data()
    fill_matrix_type_given_memory()
    fill_matrix_type_w_data()
    free_matrix_type()
    matrix_identity()
    matrix_zero()

  DESCRIPTION:
    This file contains routines for allocating, initializing and freeing
    matrices. 

*/

#include "matrix.h"

#define Min(x,y) (((x) < (y)) ? (x) : (y))
#define Max(x,y) (((x) > (y)) ? (x) : (y))

void matrix_init_row_order( m )
    matrix_type *m;
{
    int i, *p;

    if (m->row_order == NULL)
	p = m->row_order = (int *) malloc ( sizeof(int) * m->h );
    
    for (i = 0; i < m->h; i++)
	m->row_order[i] = i;
}

int matrix_size_in_bytes( w, h )
int w, h;
{
    return (sizeof (matrix_type) + sizeof(FLOATTYPE) * w * h);
}

matrix_type *matrix_malloc( w, h )
int w, h;
{
    matrix_type *m = (matrix_type *)
	malloc(matrix_size_in_bytes( w, h ));
    if( m == NULL) 
	fprintf(stderr, "malloc failed: matrix %dx%d\n", w, h);
    return m;
}

static void _fill_matrix_type( m, w, h, name )
matrix_type *m;
int w, h;
char *name;
{
    if( m != NULL) {
	m->w = w;
	m->h = h;
	m->name = name;
	m->free_data = 0;
	m->bandwidth = Max(w, h) - 1 ;
	m->row_order = NULL;
    }
}

matrix_type *copy_matrix_type( matrix )
matrix_type *matrix;
{
    matrix_type *m = matrix_malloc( matrix->w, matrix->h );
    
    if( m != NULL) {
	m->w = matrix->w;
	m->h = matrix->h;
	m->m = (FLOATTYPE *) (m + 1);
	m->name = matrix->name;
	m->free_data = 0;
	m->bandwidth = m->bandwidth;
	if (matrix->row_order) {
	    int i;
	    matrix_init_row_order( m );
	    for (i = 0; i < m->h; i++)
		m->row_order[i] = m->row_order[i];
	}
	else
	  m->row_order = NULL;
	bcopy(matrix->m, m->m, sizeof(FLOATTYPE) * m->w * m->h);
    }
    return m;
}

matrix_type *make_matrix_type( w, h, name )
int w, h;
char *name;
{
    matrix_type *m = matrix_malloc( w, h );
    
    if( m != NULL) {
	m->m = (FLOATTYPE *) (m + 1);

	_fill_matrix_type( m, w, h, name );
    }
    return m;
}

matrix_type *make_matrix_type_w_data( w, h, name, data )
int w, h;
char *name;
FLOATTYPE *data;
{
    matrix_type *m =  matrix_malloc( 0, 0 );
    
    if( m != NULL) {
	m->m = data;

	_fill_matrix_type( m, w, h, name );
    }
    return m;
}

void fill_matrix_type_given_memory( m, w, h, name )
int w, h;
char *name;
matrix_type *m;
{
    if( m != NULL) {
	m->m = (FLOATTYPE *) (m + 1);

	_fill_matrix_type( m, w, h, name );
    }
}

void fill_matrix_type_w_data( m, w, h, name, data )
matrix_type *m;
int w, h;
char *name;
FLOATTYPE *data;
{
    if( m != NULL) {
	m->m = data;

	_fill_matrix_type( m, w, h, name );
    }
}

void free_matrix_type( m )
matrix_type *m;
{
    if( m != NULL) {
	if (m->free_data && ((FLOATTYPE *)(m+1)) != m->m)
	    free(m->m);
	if (m->row_order != NULL)
	    free(m->row_order);
	free(m);
    }
}

void matrix_identity( m )
matrix_type *m;
{
    register int count;
    register FLOATTYPE *p, end_p;

    if (m->w != m->h) {
	fprintf(stderr, "matrix %s -- %dx%d cannot be identity\n",
		m->name, m->w, m->h);
	return;
    }
    
    /* zero out all entries first */
    matrix_zero( m );

    /* set main diagonal to 1.0 */
    count = m->w;
    p = m->m;
    m->bandwidth = 0;

    Duff4(count, *p++ = 1.0; p += m->w);
}

void matrix_zero( m )
matrix_type *m;
{
    register int count;
    register FLOATTYPE *p, end_p;

    /* zero out all entries */
    count = m->w * m->h;
    p = m->m;
    Duff8(count, *p++ = 0.0);
}
