/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  solve.c
  Mon Mar 11 03:49:39 EST 1991

  AUTHOR:
    Martin Friedmann

  CONTENTS:
    Min()
    Max()
    matrix_find_max_pivot()
    lu_decomp()
    forward_substitute()
    backward_substitute()

  DESCRIPTION:
    Routine for solving systems of equations with LU decomposition .
  		
    The routine lu_decomp performs dumb gaussian elimination on a matrix A
    storing both resulting matrices L and U in the space originally
    occupied by A.  The routine forward_substitute then solves Ly = b, and
    the routine backward_substitute solves the final Ux = y.  So we solve
    the system Ax = b in the following way:
 	
    WHAT	HOW
 		
    A = LU	lu_decomp(A)	  		\* A is now L & U *\
    Ly = b  	forward_substitute(A, b, y) 	\* Solve for Y *\
    Ux = y	backward_substitute(A, y, x)	\* Solve for X *\
 
*/

#include "matrix.h"

#define Min(x,y) (((x) < (y)) ? (x) : (y))
#define Max(x,y) (((x) > (y)) ? (x) : (y))

static int matrix_find_max_pivot( m, col )
int col;
matrix_type *m;
{
    FLOATTYPE max_value = 0.0;
    FLOATTYPE val, cmpval;
    int i, max_i = col;
    int end_i = Min(m->h, col+matrix_bandwidth(m)+1);

    for (i = col; i < end_i; i++) {
	val = MREF(m, m->row_order[i], col);

	cmpval = fabs(val);
	if (cmpval > max_value) {
	    max_i = i;
	    max_value = cmpval;
	}
    }

    if (max_i != col) {
	int tmp_i = m->row_order[col];
	m->row_order[col] = m->row_order[max_i];
	m->row_order[max_i] = tmp_i;
    }

    if ( max_value < SMALLFLOAT)
	return FALSE;
    return TRUE;
}

/****************************************************************************
 * lu_decomp:  Performs LU-Decomposition on Matrix m with given width and
 * 		height (w & h).  The resultant matrix U and L are both stored
 * 		in the original matrix m at the completion of the routine.
 * 		L is stored in the lower part of m and U is stored on the
 * 		diagonal and upper portion of m.
 */
int lu_decomp(m)
matrix_type *m;
{
    register int col, j, i, row, subrow, h, w;
    register FLOATTYPE *mm, multiplier, inverse_pivot;
    int n = 0;

    mm = m->m;
    h = m->h;
    w = m->w;

    matrix_init_row_order(m);

    /* for each row, zero all zeros below that diagonal point */
    for (i = 0; i < h - 1; i++) {
	register FLOATTYPE *pivot_spot;
	
	if (matrix_find_max_pivot(m, i) == FALSE)
	    return FALSE;
	
	row = m->row_order[i];

	pivot_spot = &AREF(mm,row,i,w);
	inverse_pivot = 1.0 / *pivot_spot;

	for (j = i+1; j<h; j++) {
	    register FLOATTYPE *multiplier_spot, *pivot_row;
	    subrow = m->row_order[j];

	    /* Compute multiplier for this and stash it in its spot in m */
	    multiplier_spot = &AREF(mm,subrow,i,w);
	    *multiplier_spot *= inverse_pivot;
	    multiplier = *multiplier_spot++;		/* watch it! (++) */
	    
	    pivot_row = pivot_spot + 1;
	    
	    for (col = i + 1; col < w; col++, n++)
		*multiplier_spot++ -= *pivot_row++ * multiplier;
	}
    }
    return TRUE;
}

/****************************************************************************
 * forward_substitute:	Performs forward substitution on a lower triangular
 * 		matrix with a diagonal assumed to be all ones (a matrix of the
 * 		form outputted from lu_decomp).  This routine only reads the
 * 		lower entries from the matrix L
 *
 * 		solves (Ly = b) for y given vector b and matrix L[h][w]
 */
void forward_substitute(L, b, y)
matrix_type *L;
FLOATTYPE *b, *y;
{
    register int col, i, row;
    register FLOATTYPE *rowptr;

    /* for each row i compute y[i] */
    for (i = 0; i < L->h; i++) {
	row = L->row_order[i];
	rowptr = &MREF(L, row, 0);
	y[i] = b[row];

	/* subtract out components of y already computed */
	for (col = 0; col < i; col++)
	    y[i] -= *rowptr++ * y[col];
    }
}
/****************************************************************************
 * backward_substitute:	Performs backward substitution on an upper triangular
 * 		matrix (a matrix of the form outputted from lu_decomp).  This
 * 		routine only reads the diagonal and upper entries from the
 * 		matrix U and ignores the lower values of U.
 *
 * 		solves (Ux = y) for x given vector y and matrix U[h][w]
 */
void backward_substitute(U, y, x)
matrix_type *U;
register FLOATTYPE *x, *y;
{
    register int i, col, row, h;
    register FLOATTYPE *rowptr, sum, pivot;

    h = U->h;

    /* for each row compute x[row] */
    for (i = h - 1; i >= 0; i--) {
	row = U->row_order[i];
	rowptr = &MREF(U, row, i);
	pivot = 1.0 / *rowptr++;

	/* subtract out components of x already computed */
	for (sum = 0.0, col = i + 1; col < h; col++)
	    sum += *rowptr++ * x[col];
	
 	x[i] = (y[i] - sum) * pivot;
    }
}
