/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  sub_matrix.c
  13 December 1992

  AUTHOR:
    Ali Azarbayejani

  CONTENTS:
    sub_matrix_multiply_error()
    sub_matrix_multiply()
    sub_matrix_transpose_multiply()
    sub_matrix_multiply_transpose()
    sub_matrix_multiply_vector()
    sub_matrix_add()
    sub_matrix_subtract()
    sub_matrix_scale()
    sub_matrix_invert()
    vector_normalize()

  DESCRIPTION:
    Most of this based directly on matrix.c (Martin Friedmann), modified
    to work with the sub_matrix_type.
    vector_normalize from quat.c (Martin Friedmann)
 
*/

#include "matrix.h"
#include "sub_matrix.h"

#define Min(x,y) (((x) < (y)) ? (x) : (y))
#define Max(x,y) (((x) > (y)) ? (x) : (y))

void
  sub_matrix_multiply_error(sub_matrix_type *m1,
			    sub_matrix_type *m2,
			    sub_matrix_type *m_out,
			    char *routine)
{
  fprintf(stderr, "%dx%d matrix %s X %dx%d matrix %s = %dx%d matrix %s\n",
	  m1->h, m1->w, m1->name,
	  m2->h, m2->w, m2->name,
	  m_out->h, m_out->w, m_out->name);
  fprintf(stderr, "are incompatible for %s\n", routine);
}

/****************************************************************************
 *  sub_matrix_multiply:  performs m_out = m1 X m2; 
 ****************************************************************************/
void
  sub_matrix_multiply(m1, m2, m_out)
sub_matrix_type *m1, *m2, *m_out;
{
  register int k, line1, line2, jump3, width1, width2;
  register FLOATTYPE sum, *rowptr1, *colptr2, *ptr_out;
  register int j, i;

  if ( m1->w != m2->h || m2->w != m_out->w || m1->h != m_out->h ) {
    sub_matrix_multiply_error( m1, m2, m_out, "sub_matrix_multiply" );
    return;
  }

  width1 = m1->w;
  width2 = m2->w;
  line1 = m1->line;
  line2 = m2->line;
  jump3 = m_out->line - m_out->w;
  ptr_out = m_out->m;
  
  for(i=0; i<m1->h; i++) {          /* step down  */
    for (j=0; j<width2; j++) {        /* step across  */
      sum = 0.0;
      rowptr1 = (FLOATTYPE *) (m1->m + i * line1);
      colptr2 = (FLOATTYPE *) (m2->m + j);
      k = width1;
      Duff8(k, sum += *rowptr1++ * *colptr2; colptr2 += line2);
      *ptr_out++ = sum;
    }
    ptr_out += jump3;
  }
}

/****************************************************************************
 *  sub_matrix_transpose_multiply:  performs m_out = transpose(m1) X m2; 
 ****************************************************************************/
void 
  sub_matrix_transpose_multiply(m1, m2, m_out)
sub_matrix_type *m1, *m2, *m_out;
{
  register int i, j, k, line1, line2, jump3, width1, width2, height1;
  register FLOATTYPE sum, *colptr1, *colptr2, *ptr_out;
    
  if (m1->h != m2->h || m2->w != m_out->w || m1->w != m_out->h ) {
    sub_matrix_multiply_error( m1, m2, m_out,
			      "sub_matrix_transpose_multiply" );
    return;
  }
    
  height1 = m1->h;
  width1 = m1->w;
  width2 = m2->w;
  line1 = m1->line;
  line2 = m2->line;
  jump3 = m_out->line - m_out->w;
  ptr_out = m_out->m;
    
  for (i = 0; i < width1; i++) {
    for (k = 0; k < width2; k++) {
      sum = 0.0;
      colptr1 = m1->m + i;
      colptr2 = m2->m + k;
      j = height1;
      Duff8(j, sum += *colptr1 * *colptr2; colptr1 += line1; colptr2 += line2);
      *ptr_out++ = sum;
    }
    ptr_out += jump3;
  }
}

/****************************************************************************
 *  sub_matrix_multiply_transpose:  performs m_out = m1 X transpose(m2); 
 ****************************************************************************/
void
  sub_matrix_multiply_transpose(m1, m2, m_out)
sub_matrix_type *m1, *m2, *m_out;
{
  register int i, j, k, line1, line2, jump3, width1, height1, height2;
  register FLOATTYPE sum, *rowptr1, *rowptr2, *ptr_out;
    
  if (m1->w != m2->w || m2->h != m_out->w || m1->h != m_out->h ) {
    sub_matrix_multiply_error(m1, m2, m_out, "sub_matrix_multiply_transpose" );
    return;
  }
  
  width1 = m1->w;
  height1 = m1->h;
  height2 = m2->h;
  line1 = m1->line;
  line2 = m2->line;
  jump3 = m_out->line - m_out->w;
  ptr_out = m_out->m;
  
  for (i = 0; i < height1; i++) {
    for (k = 0; k < height2; k++) {
      rowptr1 = m1->m + ( i * line1 );
      rowptr2 = m2->m + ( k * line2 );
      sum = 0.0;
      j = width1;
      Duff8( j, sum += *rowptr1++ * *rowptr2++ );
      *ptr_out++ = sum;
    }
    ptr_out += jump3;
  }
}

/****************************************************************************
 *  sub_matrix_multiply_vector:  performs v_out = m1 v ; 
 ****************************************************************************/
void
  sub_matrix_multiply_vector(m, v, v_out)
sub_matrix_type *m;
FLOATTYPE *v, *v_out;
{
  register int k, line, width;
  register FLOATTYPE sum, *rowptr, *vptr;
  register int i;

  width = m->w;
  line = m->line;
    
  for (i = 0; i < m->h; i++){          /* step down  */
    rowptr = m->m + i * line;
    sum = 0.0;
    vptr = v;
    k = width;
    Duff8( k, sum += *rowptr++ * *vptr++ );
    *v_out++ = sum;
  }
}

/****************************************************************************
 * sub_matrix_add: Adds two sub_matrices together...  
 ****************************************************************************/
void
  sub_matrix_add(m1, m2, m_out)
sub_matrix_type *m1, *m2, *m_out;
{
  register FLOATTYPE *p1, *p2, *out;
  register int i, count, width = m1->w;

#ifdef DEBUG
    if (m1->w != m2->w || m1->w != m_out->w ||
	m1->h != m2->h || m1->h != m_out->h ) {
	fprintf(stderr, "error: %s = %s + %s\n",
		m_out->name, m1->name, m2->name);
	return;
    }
#endif
  for(i=0; i<m1->h; i++) {
    count = width;
    p1 = m1->m + i * m1->line;
    p2 = m2->m + i * m2->line;
    out = m_out->m + i * m_out->line;
    Duff8( count, *out++ = *p1++ + *p2++ );
  }
}
/****************************************************************************
 * sub_matrix_subtract: subtracts two matrices together... OK if (m? == m_out) 
 ****************************************************************************/
void 
  sub_matrix_subtract(m1, m2, m_out)
sub_matrix_type *m1, *m2, *m_out;
{
  register int i, count, width = m1->w;
  register FLOATTYPE *mm1, *mm2, *out;

#ifdef DEBUG
    if (m1->w != m2->w || m1->w != m_out->w ||
	m1->h != m2->h || m1->h != m_out->h ) {
	fprintf(stderr, "error: %s = %s - %s\n",
		m_out->name, m1->name, m2->name);
	return;
    }
#endif
  for(i=0; i<m1->h; i++) {
    count = width;
    mm1 = m1->m + i * m1->line;
    mm2 = m2->m + i * m2->line;
    out = m_out->m + i * m_out->line;
    Duff8( count, *out++ = *mm1++ - *mm2++ );
  }
}
/****************************************************************************
 * sub_matrix_scale: Scales matrix by constant  OK if (m == m_out) 
 ****************************************************************************/
void
  sub_matrix_scale(m, c, m_out)
FLOATTYPE c;
sub_matrix_type *m, *m_out;
{
  register int i, count;
  register FLOATTYPE *mm1, *out;
    
  if (m->w != m_out->w || m->h != m_out->h ) {
    fprintf(stderr, "error: %s = %s * %g\n",
	    m_out->name, m->name, c);
    return;
  }
  for(i=0; i<m->h; i++) {
    count = m->w;
    mm1 = m->m + i * m->line;
    out = m_out->m + i*m_out->line;
    Duff8(count, *out++ = *mm1++ * c );
  }
}


/****************************************************************************
 * sub_matrix_invert: Performs matrix inversion using gaussian elimination.
 *		Because lu_decomp is used, the matrix m is copied for the
 * 		computation.  If you wish to obtain the LU decomposition
 * 		also, call sub_matrix_invert_lu_too or write it if I havent.
 */
void
  sub_matrix_invert( m, minverse )
sub_matrix_type *m, *minverse;
{
  matrix_type *m_in, *m_out;

  if(m->w == m->M->w && m->h == m->M->h && m->m == m->M->m)
    m_in = m->M;
  else
    m_in = copy_sub_matrix_to_matrix(m);

  if(minverse->w == minverse->M->w && minverse->h == minverse->M->h
     && minverse->m == minverse->M->m)
    m_out = minverse->M;
  else
    m_out = make_matrix_type(m->w, m->h, NULL);

  matrix_invert(m_in, m_out);

  if(m_in != m->M)
    free_matrix_type(m_in);
  if(m_out != minverse->M) {
    copy_matrix_to_sub_matrix(m_out, minverse);
    free_matrix_type(m_out);
  }
}

/************************************************************************
 * vector_normalize: okay if invector = outvector
 ************************************************************************/
int
  vector_normalize(FLOATTYPE *invector, FLOATTYPE *outvector )
{
  FLOATTYPE length;

  length = sqrt ( vector_dot3(invector, invector) );

  if ( (length < 0.00000001) && (length > -0.00000001) ) {
    outvector[0] = outvector[1] = outvector[2] = 0.0;
  }
  else {
    outvector[0] = invector[0] / length;
    outvector[1] = invector[1] / length;
    outvector[2] = invector[2] / length;
  }
}

