/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  sub_matrix_init.c
  13 December 1992

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    flex_matrix_size_in_bytes()
    flex_matrix_malloc()
    _fill_sub_matrix_type()
    fill_flex_matrix_type_given_memory()
    make_flex_matrix_type()
    fill_sub_matrix_type_given_matrix()
    make_sub_matrix_type()
    copy_sub_matrix_type()
    copy_sub_matrix_to_matrix()
    copy_matrix_to_sub_matrix()
    sub_matrix_identity()
    sub_matrix_zero()
    sub_matrix_reframe()

  DESCRIPTION:
    Initialize sub-matrix.

*/

#include "matrix.h"
#include "sub_matrix.h"

#define Min(x,y) (((x) < (y)) ? (x) : (y))
#define Max(x,y) (((x) > (y)) ? (x) : (y))

int
  flex_matrix_size_in_bytes(int w, int h)
{
  return(sizeof(sub_matrix_type) + matrix_size_in_bytes(w, h));
}

flex_matrix_type *
  flex_matrix_malloc(int w, int h)
{
  flex_matrix_type *m = (flex_matrix_type *)
    malloc(flex_matrix_size_in_bytes( w, h ));
  if( m == NULL) 
    fprintf(stderr, "malloc failed: flex_matrix %dx%d\n", w, h);
  return m;
}

static void
  _fill_sub_matrix_type(sub_matrix_type *m, int w, int h, int l, char *name)
{
  if( m != NULL) {
    m->w = w;
    m->h = h;
    m->name = name;
    m->line = l;
  }
}

void 
  fill_flex_matrix_type_given_memory(flex_matrix_type *m,
				     int w,
				     int h,
				     char *name)
{
  if( m != NULL) {
    m->M = (matrix_type *) (m + 1);
    /* Can't do this!  It is not guaranteed that a matrix_type */
    /* can be aligned to sub_matrix_type-size boundaries */
    /* THIS IS A BUG!!*/
    fill_matrix_type_given_memory(m->M, w, h, name);

    m->m = m->M->m;
    _fill_sub_matrix_type(m, w, h, m->M->w, name);
  }
}

flex_matrix_type *
  make_flex_matrix_type(int w, int h, char *name)
{
  flex_matrix_type *m = flex_matrix_malloc(w, h);

  fill_flex_matrix_type_given_memory(m, w, h, name);
  return m;
}

void
  fill_sub_matrix_given_matrix(sub_matrix_type *sm,
			       matrix_type *M,
			       int r, 
			       int c,
			       int h,
			       int w, 
			       char *name)
{
  if(NULL != sm) {
    sm->M = M;
    sm->w = w;
    sm->h = h;
    sm->line = M->w;
    sm->name = name;
    sm->m = M->m + c * M->w + r;
  }
}


sub_matrix_type *
  make_sub_matrix_type(matrix_type *M, int r, int c, int h, int w, char *name)
{
  sub_matrix_type *m = (sub_matrix_type *) malloc(sizeof(sub_matrix_type));

  fill_sub_matrix_given_matrix(m, M, r, c, h, w, name);
}

void 
  copy_sub_matrix_type(sub_matrix_type *src, sub_matrix_type *dst)
{
  register int count;
  register FLOATTYPE *src_p, *dst_p;

  if(src->w != dst->w || src->h != dst->h)
    return;

  count = dst->h;
  src_p = src->m;
  dst_p = dst->m;
  while(count--) {
    bcopy(src_p, dst_p, sizeof(FLOATTYPE)*src->w);
    src_p += src->line;
    dst_p += dst->line;
  }
}
  
matrix_type *
  copy_sub_matrix_to_matrix(sub_matrix_type *sm)
{
  matrix_type *M = make_matrix_type(sm->w, sm->h, sm->name);
  register FLOATTYPE *src_p, *dst_p;
  register int count;
    
  if(NULL != M) {
    count = sm->h;
    src_p = sm->m;
    dst_p = M->m;
    while(count--) {
      bcopy(src_p, dst_p, sizeof(FLOATTYPE)*sm->w);
      src_p += sm->line;
      dst_p += M->w;
    }
  }
  return M;
}

void
  copy_matrix_to_sub_matrix(matrix_type *M, sub_matrix_type *sm)
{
  register FLOATTYPE *src_p, *dst_p;
  register int count;
    
  if(M->w != sm->w || M->h != sm->h)
    return;

  count = sm->h;
  src_p = M->m;
  dst_p = sm->m;
  while(count--) {
    bcopy(src_p, dst_p, sizeof(FLOATTYPE)*M->w);
    src_p += M->w;
    dst_p += sm->line;
  }
}

void
  sub_matrix_identity(sub_matrix_type *sm)
{
  register int count;
  register FLOATTYPE *p;

    if (sm->w != sm->h) {
      fprintf(stderr, "sub_matrix %s -- %dx%d cannot be identity\n",
	      sm->name, sm->w, sm->h);
      return;
    }
    
    sub_matrix_zero(sm);

    count = sm->w;
    p = sm->m;
    Duff4(count, *p++ = 1.0; p += sm->line);
}

void 
  sub_matrix_zero(sub_matrix_type *sm)
{
  register int count, len;
  register FLOATTYPE *p, *v;

  count = sm->h;
  p = sm->m;
  while(count--) {
    len = sm->w;
    v = p;
    Duff8(len, *v++ = 0.0);
    p += sm->line;
  }
}

void
  sub_matrix_reframe(sub_matrix_type *m, int r, int c, int h, int w)
{
  /* could do boundary checking here...*/
  if(NULL != m) {
    m->w = w;
    m->h = h;
    m->m = m->M->m + c * m->M->w + r;
  }
}
