/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  est.c
  26 May 1993

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    est_error()
    est_warning()
    est_list_init()
    est_list_expand()
    est_list_check()
    est_list_input()
    est_get_alpha()
    est_put_alpha()
    est_scan()
    est_print()
    est_close()
    est_open_read()
    est_open_write()

  DESCRIPTION:
    State estimate data for motion, pointwise structure, and camera

*/

#include <stdio.h>
#include <stdlib.h>
#include "files.h"

               /********************************\
----------------------------------------------------------------
   Utility functions
----------------------------------------------------------------
               \********************************/

static void
  est_error(char *s)
{
  fprintf(stderr, "est ERROR: %s\n", s);
  exit(0);
}

static void
  est_warning(char *s)
{
  fprintf(stderr, "est WARNING: %s\n", s);
}

               /********************************\
----------------------------------------------------------------
   List handlers.
   These can be taken out and used in general for maintaining
    lists of things.
----------------------------------------------------------------
               \********************************/

static void *
  est_list_init(size_t obj_size, int *n_total_p)
{
  void * ptr = malloc(FILES_LIST_INIT * obj_size);

  if(NULL == ptr)
    est_error("est_list_init: failed to malloc");

  *n_total_p = FILES_LIST_INIT;
  return ptr;
}

static void *
  est_list_expand(size_t obj_size, void *list, int n_used, int* n_total_p)
{
  void * ptr;

  if(obj_size < 0 || NULL == list || n_used < 0 || *n_total_p <= 0
     || n_used > *n_total_p)
    est_error("est_list_expand: stupid args");

  ptr = malloc( (*n_total_p + FILES_LIST_INC) * obj_size );

  if(NULL == ptr)
    est_error("est_list_expand: malloc failed");

  bcopy(list, ptr, n_used * obj_size);
  free(list);

  *n_total_p += FILES_LIST_INC;
  return ptr;
}

/* Outputs valid "list" and "*n_total_p" */
static void *
  est_list_check(void *list, int n_used, int *n_total_p, size_t obj_size)
{
  if(list && n_used < *n_total_p)
    return list;

  else if(NULL == list && 0 == n_used && 0 == *n_total_p && obj_size > 0)
    return est_list_init(obj_size, n_total_p);

  else if(list && n_used == *n_total_p && obj_size > 0)
    return est_list_expand(obj_size, list, n_used, n_total_p);

  else
    est_error("est_list_check: uh-oh, error in args?");
}

static void
  est_list_input(void *in, void **l_p, int *nu_p, int *nt_p, size_t obj_size)
{
  void *ptr;

  *l_p = est_list_check(*l_p, *nu_p, nt_p, obj_size);
  
  ptr = (char *)(*l_p) + (*nu_p) * obj_size;
  bcopy(in, ptr, obj_size);
  (*nu_p)++;
}

               /********************************\
----------------------------------------------------------------
   External functions.
----------------------------------------------------------------
               \********************************/

static alpha_type *
  est_get_alpha(est_type *E)
{
  if(FILES_OPEN_READ == E->mode) {
    if(E->N > 0)
      return( E->alphas + --E->N );
    else
      return ((void *)EOF);
  }
  else
    est_warning("est_get_alpha: file not open for read");
}

static void
  est_put_alpha(est_type *E, alpha_type *ap)
{
  if(FILES_OPEN_WRITE == E->mode)
    est_list_input((void *)ap, (void **)&E->alphas,
		   &E->N, &E->alphas_size, sizeof(alpha_type));
  else 
    est_warning("est_put_alpha: file not open for write");
}

static int
  est_scan(est_type *E)
{
  int count, nconv, total = 0;
  alpha_type at;

  if(E->mode != FILES_OPEN_READ) {
    est_warning("est_scan: File not open for read");
    return EOF;
  }

  nconv = fscanf(E->fp, "%g%g%g", &E->x, &E->y, &E->zb);
  if(EOF == nconv)
    return EOF;
  else if(3 != nconv)
    est_error("est_scan: Read error position");

  if(4 != fscanf(E->fp, "%g%g%g%g", &E->q0, &E->q1, &E->q2, &E->q3))
    est_error("est_scan: Read error quaternion");

  if(1 != fscanf(E->fp, "%g", &E->b))
    est_error("est_scan: Read error beta");
  
  if(2 != fscanf(E->fp, "%g%g", &E->xp, &E->yp))
    est_error("est_scan: Read error principal point");
  
  E->N = 0;
  if(1 != fscanf(E->fp, "%d", &count))
    est_error("est_scan: Read error alphas");
  total += count;
  while(count--) {
    if(2 != fscanf(E->fp, "%d%g", &at.id, &at.a))
      est_error("est_scan: Read error alpha");
    est_list_input((void *)&at, (void **)&E->alphas,
		   &E->N, &E->alphas_size, sizeof(alpha_type));
  }
  return total;
}

static void
  est_print(est_type *E)
{
  alpha_type *ap;

  if(NULL != E->fp && FILES_OPEN_WRITE == E->mode && E->N >=0) {

    fprintf(E->fp, "%.10g %.10g %.10g\n", E->x, E->y, E->zb);
    fprintf(E->fp, "%.10g %.10g %.10g %.10g\n", E->q0, E->q1, E->q2, E->q3);
    fprintf(E->fp, "%.10g\n", E->b);
    fprintf(E->fp, "%.10g %.10g\n", E->xp, E->yp);

    fprintf(E->fp, "%d\n", E->N);
    while(E->N) {
      ap = E->alphas + --E->N;
      fprintf(E->fp, "  %d %.10g\n", ap->id, ap->a);
    }
    fprintf(E->fp, "\n");
  }
  else
    est_error("est_print: bad args");
}

static void
  est_close(est_type *E)
{
  if(FILES_CLOSED == E->mode) {
    est_warning("est_close: file aint open");
    return;
  }
  E->filename = NULL;
  fclose(E->fp);
  E->fp = NULL;
  E->mode = FILES_CLOSED;
}

est_type *
  est_open_read(char *name)
{
  FILE *fp;
  register est_type *E = NULL;

  fp = (NULL == name) ? stdin : fopen(name, "r");
#ifdef VX
  if(1) {
#else
  if(NULL != fp) {
#endif
    E = (est_type *) malloc(sizeof(est_type));
    if(NULL != E) {

      E->N = 0;
      E->alphas = NULL;
      E->alphas_size = 0;

      E->filename = name;
      E->fp = fp;
      E->mode = FILES_OPEN_READ;

      E->get_alpha = est_get_alpha;
      E->put_alpha = est_put_alpha;
      E->scan = est_scan;
      E->print = est_print;
      E->close = est_close;
    }
    else
      fclose(fp);
  }
  return E;
}

est_type *
  est_open_write(char *name)
{
  FILE *fp;
  register est_type *E = NULL;

  fp = (NULL == name) ? stdout : fopen(name, "w");
  if(NULL != fp) {
    E = (est_type *) malloc(sizeof(est_type));
    if(NULL != E) {

      E->x = E->y = E->zb = 0.;
      E->q0 = 1.;
      E->q1 = E->q2 = E->q3 = 0.;
      E->b = 1.;
      E->xp = E->yp = 0.;

      E->N = 0;
      E->alphas = NULL;
      E->alphas_size = 0;

      E->filename = name;
      E->fp = fp;
      E->mode = FILES_OPEN_WRITE;

      E->get_alpha = est_get_alpha;
      E->put_alpha = est_put_alpha;
      E->scan = est_scan;
      E->print = est_print;
      E->close = est_close;
    }
    else
      fclose(fp);
  }
  return E;
}
