/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  ftstoraw.c
  7 June 1993

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    Min()
    Max()
    Limit()
    ParseCmdLine()
    draw_hline()
    draw_vline()
    draw_box()
    draw_feature()
    clear_image()
    main()

  DESCRIPTION:
    Makes a raw image from a fts file with a small box where each feature
    point lies.

*/

#include <stdio.h>
#include <stdlib.h>
#include "files.h"

#define Min(a,b) (((a)<(b)) ? (a) : (b))
#define Max(a,b) (((a)>(b)) ? (a) : (b))
#define Limit(x, min, max)  Min(max, Max(min, x))

static struct {
  int	verbose;
  unsigned int color;
  int	width;
  int	height;
  int	dbox;
  float	min, max;
  int	single;
  int	depthcue;
  FILE	*imagefile;
}
params = { 0, 255, 256, 240, 2, 0., 50., -1, 0, NULL };

static void
  ParseCmdLine(int argc, char **argv)
{
  int		i;
  char		usage[1024], io[128], *arg;

  sprintf(usage, "%s", "[-v] [-w #] [-h #] [-single FEAT_ID] [-image <file>] [-depthcue]");
  sprintf(io,"\n\t%s", "< (.fts) > (.raw)");

  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i], "-v"))
      params.verbose = 1;
    else if(!strcmp(argv[i], "-c"))
      params.color = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-w"))
      params.width = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-h"))
      params.height = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-dbox"))
      params.dbox = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-min"))
      params.min = atof(argv[++i]);
    else if(!strcmp(argv[i], "-max"))
      params.max = atof(argv[++i]);
    else if(!strcmp(argv[i], "-single"))
      params.single = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-depthcue"))
      params.depthcue = 1;
    else if(!strcmp(argv[i], "-image"))
      params.imagefile = fopen(argv[++i], "r");
    else {
      fprintf(stderr,
	      "usage: %s %s %s\n",
	      argv[0], usage, io);
      exit(1);
    }
  }
}

void
  draw_hline(unsigned char *buffer, int line,
	     int row, int col, int width, unsigned char value)
{
  unsigned char *ucp;
  int	j;

  ucp = buffer + row*line + col;
  for(j=0; j<width; j++)
    *ucp++ = value;
}

void
  draw_vline(unsigned char *buffer, int line,
	     int row, int col, int height, unsigned char value)
{
  unsigned char *ucp;
  int	i;

  ucp = buffer + row*line + col;
  for(i=0; i<height; i++) {
    *ucp = value;
    ucp += line;
  }
}


void
  draw_box(unsigned char *buffer, int line,
	   int row, int col, int dx, unsigned char value)
{
  unsigned char		*ucp;
  int		length;

  length = 2*dx + 1;
  draw_hline(buffer, line, row-dx, col-dx, length, value);
  draw_hline(buffer, line, row+dx, col-dx, length, value);
  draw_vline(buffer, line, row-dx, col-dx, length, value);
  draw_vline(buffer, line, row-dx, col+dx, length, value);
}

#define DBOX 2

void
  draw_feature(float x, float y, unsigned char *image, float depth)
{
  int 	X_pixels, Y_pixels;
  int	pixels_per_pic = Max(params.width, params.height);
  float intensity;
  unsigned char uint = params.color;

  X_pixels = x * pixels_per_pic + params.width / 2;
  X_pixels = Limit(X_pixels, params.dbox, params.width - 1 - params.dbox);

  Y_pixels = y * pixels_per_pic + params.height / 2;
  Y_pixels = Limit(Y_pixels, params.dbox, params.height - 1 - params.dbox);

#if 0
  fprintf( stderr, "xy= %f, %f\n", x, y );
  fprintf( stderr, "\tXY= %d, %d\n", X_pixels, Y_pixels );
  fprintf( stderr, "\tpixels_per_pic= %f\n", pixels_per_pic );
#endif

  if( params.depthcue ) {
    intensity = (params.max - depth) * 255. / (params.max - params.min);
    uint =   (unsigned char)Limit( intensity, 0., 255.);
  }
#if 0
  fprintf(stderr, "max= %f, min= %f, depth= %f, int= %f, uint= %d\n",
	  params.max, params.min, depth, intensity, uint);
#endif
  draw_box(image, params.width, Y_pixels, X_pixels, params.dbox, uint);
	 
}

void
  clear_image(unsigned char *image)
{
  register int count = params.width * params.height;

  while(count--)
    *image++ = 0;
}


main(int argc, char **argv)
{
  fts_type	*Feat;
  unsigned char	*image = NULL;
  new_type	*np;
  old_type	*op;

  ParseCmdLine(argc, argv);

  Feat = fts_open_read(NULL);
  image = (unsigned char *) malloc(params.width * params.height);
  if( !image ) {
    fprintf( stderr, "ftstoraw: failed to allocate image\n" );
    exit(1);
  }

  while(EOF != Feat->scan(Feat)) {
    if(NULL != params.imagefile)
      fread(image, sizeof(char), params.width*params.height, params.imagefile);
    else
      clear_image(image);

    if(params.single < 0) { /* Do all features */
      while(EOF != (int)(np = Feat->get_new(Feat)))
	draw_feature(np->x, np->y, image, 0.);
      while(EOF != (int)(op = Feat->get_old(Feat)))
	draw_feature(op->x, op->y, image, op->Vx);
    }
    else { /* Single feature */
      int found_id = 0;

      while(EOF != (int)(np = Feat->get_new(Feat)))
	if(np->id == params.single) {
	  found_id = 1;
	  fprintf(stderr, "Feature %d inited as new\n", params.single);
	  draw_feature(np->x, np->y, image, 0.);
	  break;
	}
      if(!found_id)
	while(EOF != (int)(op = Feat->get_old(Feat)))
	  if(op->id == params.single) {
	    found_id = 1;
	    draw_feature(op->x, op->y, image, 0.);
	    break;
	  }
    }
    fwrite(image, sizeof(char), params.width*params.height, stdout);
  }
  exit(0);
}
