/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  protoest.c
  7 June 1993

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    typedef motion_state_type
    typedef motion_args_type
    typedef motion_cmd_type
    error()
    my_quat_multiply()
    my_quat_rotate()
    my_vector_dot()
    my_vector_normalize()
    motion_cmd_rotate()
    motion_cmd_rotate_about()
    motion_cmd_trans()
    motion_cmd_make()
    motion_cmd_scan()
    typedef motion_spec_type
    motion_spec_scan()
    motion_spec_step()
    motion_spec_make()
    parse_command_line()
    main()

  DESCRIPTION:
    Takes a motion profile file and computes a .est file representing
    the motion.  An example set of profile commands:
    ----------------
    INIT_TRANSLATION 0 0 2
    INIT_QUATERNION 1 0 0 0
    BETA 1
    PPT 0. 0.
    frames 90
      rotate 2 deg 0 1 0
    end
    ----------------

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "matrix/matrix.h"
#include "files.h"

#ifndef M_PI
#define M_PI       3.1415926535897932385E0  /*Hex  2^ 1 * 1.921FB54442D18 */
#endif
#define	DT	1
#define RADS_PER_DEG (M_PI / 180.)
#define FLOATTYPE float

typedef struct {
  FLOATTYPE	t[3];
  FLOATTYPE	q[4];
  FLOATTYPE	b;
  FLOATTYPE	ppt[2];
} motion_state_type;

typedef union {
  struct { FLOATTYPE ang, rads, axis[3]; } rotate;
  struct { FLOATTYPE pt[3], ang, rads, axis[3]; } rotate_about;
  struct { FLOATTYPE t[3]; } trans;
} motion_args_type;

typedef struct _motion_cmd_type {
  void			(*fn)();
  motion_args_type	args;
} motion_cmd_type;

static void
  error(char *s)
{
  fprintf(stderr, "protoest: ERROR: %s\n", s);
  exit(1);
}

static void
  my_quat_multiply(FLOATTYPE  *q1, FLOATTYPE  *q2, FLOATTYPE  *q3)
{
  FLOATTYPE              qt[4];

  if(NULL == q3 || q1 == q3 || q2 == q3) {
    qt[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3];
    qt[1] = q1[1]*q2[0] + q1[0]*q2[1] - q1[3]*q2[2] + q1[2]*q2[3];
    qt[2] = q1[2]*q2[0] + q1[3]*q2[1] + q1[0]*q2[2] - q1[1]*q2[3];
    qt[3] = q1[3]*q2[0] - q1[2]*q2[1] + q1[1]*q2[2] + q1[0]*q2[3];
    if(q3) {
      q3[0] = qt[0];
      q3[1] = qt[1];
      q3[2] = qt[2];
      q3[3] = qt[3];
    }
  }
  else {
    q3[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3];
    q3[1] = q1[1]*q2[0] + q1[0]*q2[1] - q1[3]*q2[2] + q1[2]*q2[3];
    q3[2] = q1[2]*q2[0] + q1[3]*q2[1] + q1[0]*q2[2] - q1[1]*q2[3];
    q3[3] = q1[3]*q2[0] - q1[2]*q2[1] + q1[1]*q2[2] + q1[0]*q2[3];
  }
}

static void
  my_quat_rotate(FLOATTYPE *q, FLOATTYPE *in, FLOATTYPE *out)
{
  FLOATTYPE q01, q02, q03, q11, q12, q13, q22, q23, q33, *p;
  FLOATTYPE r[9], inbuf[3];

  vector_copy(in, inbuf, 3);

  q01 = q[0]*q[1];
  q02 = q[0]*q[2];
  q03 = q[0]*q[3];
  q11 = q[1]*q[1];
  q12 = q[1]*q[2];
  q13 = q[1]*q[3];
  q22 = q[2]*q[2];
  q23 = q[2]*q[3];
  q33 = q[3]*q[3];
  
  p = r;
  *p++ = 1.0 - 2 * (q22 + q33);
  *p++ =     - 2 * (q03 - q12);
  *p++ =       2 * (q02 + q13);
  *p++ =       2 * (q03 + q12);
  *p++ = 1.0 - 2 * (q11 + q33);
  *p++ =     - 2 * (q01 - q23);
  *p++ =     - 2 * (q02 - q13);
  *p++ =       2 * (q01 + q23);
  *p++ = 1.0 - 2 * (q11 + q22);

  out[0] = r[0]*inbuf[0] + r[1]*inbuf[1] + r[2]*inbuf[2];
  out[1] = r[3]*inbuf[0] + r[4]*inbuf[1] + r[5]*inbuf[2];
  out[2] = r[6]*inbuf[0] + r[7]*inbuf[1] + r[8]*inbuf[2];
}

static FLOATTYPE
  my_vector_dot(FLOATTYPE *v1, FLOATTYPE *v2, int n)
{
  register FLOATTYPE sum = 0.;

  while(n--)
    sum += (*v1++) * (*v2++);
  return sum;
}

static void
  my_vector_normalize(FLOATTYPE *invector, FLOATTYPE *outvector, int n )
{
  FLOATTYPE length;

  length = sqrt ( my_vector_dot(invector, invector, n) );

  if ( (length < 0.00000001) && (length > -0.00000001) )
    vector_zero(outvector, n);
  else
    while(n--)
      outvector[n] = invector[n] / length;
}

static void
  motion_cmd_rotate(motion_args_type *args,
		    motion_state_type *state)
{
  FLOATTYPE ang, cost2, sint2, dq[4];

  ang = args->rotate.ang * args->rotate.rads;
  vector_copy(args->rotate.axis, dq+1, 3);
  my_vector_normalize(dq+1, dq+1, 3);

  dq[0] = cos( (double)ang / 2. );
  sint2 = sin( (double)ang / 2. );
  vector_scale3(dq+1, sint2, dq+1);

  my_quat_multiply(state->q, dq, state->q);
  my_vector_normalize(state->q, state->q, 4);
}

static void
  motion_cmd_rotate_about(motion_args_type *args,
			  motion_state_type *state)
{
  FLOATTYPE ang, cost2, sint2, dq[4], dt[3], sum[3];

  /* R = RR* */
  ang = args->rotate_about.ang * args->rotate_about.rads;
  vector_copy(args->rotate_about.axis, dq+1, 3);
  my_vector_normalize(dq+1, dq+1, 3);
  dq[0] = cos( (double)ang / 2. );
  sint2 = sin( (double)ang / 2. );
  vector_scale3(dq+1, sint2, dq+1);
  my_vector_normalize(dq, dq, 4);
  my_quat_multiply(state->q, dq, state->q);
  my_vector_normalize(state->q, state->q, 4);

  /* t = R* (t - dt*) + dt* */
  vector_copy(args->rotate_about.pt, dt, 3);
  vector_subtract3(state->t, dt, sum);
  my_quat_rotate(dq, sum, sum);
  vector_add3(sum, dt, state->t);
}

static void
  motion_cmd_trans(motion_args_type *args,
		   motion_state_type *state)
{
  vector_add3(state->t, args->trans.t, state->t);
}

static motion_cmd_type *
  motion_cmd_make()
{
  void *cmd = malloc(sizeof(motion_cmd_type));

  if(cmd)
    return( (motion_cmd_type *) cmd);
  else
    return NULL;
}

static motion_cmd_type *
  motion_cmd_scan(motion_cmd_type *M, FILE *stream)
{
  char			cmd[128];
  char			units[5];

  fscanf(stream, "%s", cmd);

  if( !strcmp(cmd, "end") )
    M = NULL;
  
  else if( !strcmp(cmd, "rotate") ) {
    M->fn = motion_cmd_rotate;
    fscanf(stream, "%f", &M->args.rotate.ang);
    fscanf(stream, "%s", units);
    if(!strcmp(units, "deg"))
      M->args.rotate.rads = RADS_PER_DEG;
    else if(!strcmp(units, "rad"))
      M->args.rotate.rads = 1.;
    else
      error(units);
    fscanf(stream, "%f%f%f",
	   &M->args.rotate.axis[0],
	   &M->args.rotate.axis[1],
	   &M->args.rotate.axis[2]);
  }

  else if( !strcmp(cmd, "rotate_about") ) {
    M->fn = motion_cmd_rotate_about;
    fscanf(stream, "%f%f%f",
	   &M->args.rotate_about.pt[0],
	   &M->args.rotate_about.pt[1],
	   &M->args.rotate_about.pt[2]);
    fscanf(stream, "%f", &M->args.rotate_about.ang);
    fscanf(stream, "%s", units);
    if(!strcmp(units, "deg"))
      M->args.rotate_about.rads = RADS_PER_DEG;
    else if(!strcmp(units, "rad"))
      M->args.rotate_about.rads = 1.;
    else
      error(units);
    fscanf(stream, "%f%f%f",
	   &M->args.rotate_about.axis[0],
	   &M->args.rotate_about.axis[1],
	   &M->args.rotate_about.axis[2]);
  }

  else if( !strcmp(cmd, "trans") ) {
    M->fn = motion_cmd_trans;
    fscanf(stream, "%f%f%f",
	   &M->args.trans.t[0],
	   &M->args.trans.t[1],
	   &M->args.trans.t[2]);
  }

  else
    error(cmd);

  return M;
}

typedef struct {
  motion_state_type	state;
  int			nfr;	/* number of frames for this motion */
  int			ncmds;	/* number of motion commands per frame */
  motion_cmd_type	cmds[10];

  int			(*scan)();
  int			(*step)();
} motion_spec_type;

int
  motion_spec_scan(motion_spec_type *M, FILE *stream)
{
  char 	token[128];
  int rv;

  rv = fscanf(stream, "%s%d", token, &M->nfr);
  if(2 == rv) {
    if(strcmp(token, "frames"))
      error(token);
  
    M->ncmds = 0;
    while( motion_cmd_scan(M->cmds + M->ncmds, stream) )
      M->ncmds++;
  }
  return rv;
}
    
int
  motion_spec_step(motion_spec_type *M)
{
  register int count;

  if(M->nfr--) {
    for(count=0; count<M->ncmds; count++)
      M->cmds[count].fn(&M->cmds[count].args, &M->state);
    return 1;
  }
  else
    return EOF;
}

static motion_spec_type *
  motion_spec_make()
{
  motion_spec_type *M = NULL;

  M = (motion_spec_type *) malloc(sizeof(motion_spec_type));

  if(M) {
    M->state.t[0] = 0.;
    M->state.t[1] = 0.;
    M->state.t[2] = 0.;
    M->state.q[0] = 1.;
    M->state.q[1] = 0.;
    M->state.q[2] = 0.;
    M->state.q[3] = 0.;
    M->state.b = 1.;
    M->state.ppt[0] = 0.;
    M->state.ppt[1] = 0.;
    M->nfr = 0;
    M->ncmds = 0;

    M->scan = motion_spec_scan;
    M->step = motion_spec_step;
  }
  return M;
}

struct params {
  char		verbose;
}
params = { 0 };

void parse_command_line(int argc, char **argv)
{
  int		i;

  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i],"-v"))
      params.verbose = 1;
    else {
      fprintf(stderr,
	      "usage: %s %s\n",
	      argv[0],
	      "[-v]");
      exit(1);
    }
  }
}

main(int argc, char **argv)
{
  FLOATTYPE	acc[6], vel[6], pos[3], ang[3], dq[4];
  FLOATTYPE	dummy[6];
  int		nframes, n;
  est_type	*Traj;
  motion_spec_type	*Motion;

  parse_command_line(argc, argv);

  Motion = motion_spec_make();

  /* All we care about dynamically is pos/quat */
  Traj = est_open_write(NULL);
  scanf("%*s%f%f%f", &Traj->x, &Traj->y, &Traj->zb);
  scanf("%*s%f%f%f%f", &Traj->q0, &Traj->q1, &Traj->q2, &Traj->q3);
  scanf("%*s%f", &Traj->b);
  scanf("%*s%f%f", &Traj->xp, &Traj->yp);
#if 0
  vector_copy(&Traj->x, &Motion->state.t, 3);
  vector_copy(&Traj->q0, &Motion->state.q, 3);
  Motion->state.b = Traj->b;
  vector_copy(&Traj->xp, &Motion->state.ppt, 2);
#else
  vector_copy(&Traj->x, Motion->state.t, 3);
  vector_copy(&Traj->q0, Motion->state.q, 3);
  Motion->state.b = Traj->b;
  vector_copy(&Traj->xp, Motion->state.ppt, 2);
#endif
  Traj->zb *= Traj->b;
  Traj->print(Traj);

#if 0
  while(!(EOF == fscanf(stdin,
			"%d %f %f %f %f %f %f",
			&nframes,
			acc, acc+1, acc+2, acc+3, acc+4, acc+5))) {
#endif
  while( EOF != Motion->scan(Motion, stdin) ) {
    /* Have a new block of frames */
    while( EOF != Motion->step(Motion) ) {
      /* Have a new state vector */

      vector_copy( Motion->state.t, &Traj->x, 10 );
      Traj->zb *= Traj->b;
      Traj->print(Traj);
    }
  }
  exit(0);
}
