/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  se.c
  9 October 1992

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    ParseCmdLine()
    se_error()
    se_print_sta()
    se_get_initial_data()
    se_get_known_motion()
    main()

  DESCRIPTION:
    State estimator.
    Recursive estimator for motion, pointwise structure, and camera
    parameters from point correspondences in an image sequence, based
    on the extended Kalman filter.
    Takes in features, outputs state estimates.

  REFERENCES:
    Azarbayejani, Ali and Alex Pentland, "Recursive Estimation of Motion,
       Structure, and Focal Length", IEEE PAMI, June 1995

    Brown, Robert G., << Introduction to Random Signal Analysis and Kalman
        Filtering >>, John Wiley & Sons, New York, 1983

    Gelb, Arthur (Editor), << Applied Optimal Estimation >>, MIT Press,
       Cambridge, MA, 1974

    Friedmann, Martin, Alex Pentland, and Thad Starner, "Device
       Synchronization using an Optimal Linear Filter", Proceedings of the
       ACM Symposium on Interactive 3D Graphics, Cambridge, MA, Vol 26, No
       2, March 1992

*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "matrix/matrix.h"
#include "ekf/ekf.h"
#include "smkf.h"
#include "files.h"

static struct {
  int		verbose;
  char		*name;
  int		width;
  int		height;
  FLOATTYPE	b;
  FLOATTYPE	Pb;
  int		nfr;
  int		nfe;
  int		covariance;	/* [Stan] output covariance matrix */
  char		*motionfile;	/* [Trevor] input motion .est file */
  int		seq_iter;	/* iteration number of whole SEQUENCE	*/
  FLOATTYPE	Qr;
  FLOATTYPE	Qth;
  FLOATTYPE	Ro;
}
params = { 0, "se", 256, 240, -1.0, -1.0, -1, -1, 0, NULL, 0,
	     SMKF_DEFAULT_VAR_TR_NOISE,	SMKF_DEFAULT_VAR_ROT_NOISE,
	     SMKF_DEFAULT_VAR_MEAS_NOISE };

static void
  ParseCmdLine(int argc, char **argv)
{
  int		i;
  char		usage[1024], io[128], *arg;

  sprintf(usage,"%s%s",
	  "[-v] [-name <name>] [-w #] [-h #] [-b #] [-Pb #]",
	  "[-nfr #] [-nfe #] [-cov] [-motion <name>] [-Ro #]");
  sprintf(io,"\n\t%s", "< (.fts) > (.est)");

  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i], "-v"))
      params.verbose = 1;
    else if(!strcmp(argv[i], "-name"))
      params.name = argv[++i];
    else if(!strcmp(argv[i], "-w"))
      params.width = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-h"))
      params.height = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-b"))
      sscanf(argv[++i], "%f", &params.b);
    else if(!strcmp(argv[i], "-Pb"))
      sscanf(argv[++i], "%f", &params.Pb);
    else if(!strcmp(argv[i], "-nfr"))
      params.nfr = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-nfe"))
      params.nfe = atoi(argv[++i]);
    else if(!strcmp(argv[i], "-cov"))
      params.covariance = 1;
    else if(!strcmp(argv[i], "-motion"))
      params.motionfile = argv[++i];
    else if(!strcmp(argv[i], "-Qr"))
      sscanf(argv[++i], "%f", &params.Qr);
    else if(!strcmp(argv[i], "-Qth"))
      sscanf(argv[++i], "%f", &params.Qth);
    else if(!strcmp(argv[i], "-Ro"))
      sscanf(argv[++i], "%f", &params.Ro);
    else {
      fprintf(stderr,
	      "usage: %s %s %s\n",
	      argv[0], usage, io);
      exit(1);
    }
  }
}

void
  se_error(char *s)
{
  fprintf(stderr, "SE_ERROR: %s\n", s);
  exit(0);
}

void
  se_print_sta(smkf_type *F)
{
  sta_type	*Sta;
  depth_type	dt;
  int		n_alphas = F->n_feat, ekfXloc;
  char		filename[128];

  sprintf(filename, "%s.sta", params.name);
  if(NULL == (Sta = sta_open_write(filename)))
    se_error("couldn't open output .sta file");

  Sta->iter = params.seq_iter;
  Sta->b = SMKF_B(F->ekf->X_e);
  Sta->Vb = SMKF_VAR_BETA(F->ekf->P_e);
  Sta->xp = SMKF_PX(F->ekf->X_e);
  Sta->yp = SMKF_PY(F->ekf->X_e);
  Sta->Vxp = SMREF(F->ekf->P_e, 7, 7);
  Sta->Vyp = SMREF(F->ekf->P_e, 8, 8);
  Sta->dz = 0.;

  for(ekfXloc = 0; ekfXloc < n_alphas; ekfXloc++) {
    dt.id = F->Xid[ ekfXloc ];
    dt.a = SMKF_ALPHA(F->ekf->X_e, ekfXloc);
    dt.Va = SMKF_VAR_ALPHA(F->ekf->P_e, ekfXloc);
    dt.x = F->yo[2*ekfXloc];
    dt.y = F->yo[2*ekfXloc + 1];
    Sta->put_depth(Sta, &dt);
  }
  Sta->print(Sta);
  Sta->close(Sta);
}
    
void
  se_get_initial_data(fts_type 	*Fts,
		      int	n,
		      FLOATTYPE	**yo,
		      FLOATTYPE	**ao,
		      FLOATTYPE	**Pao,
		      FLOATTYPE	**bo,
		      FLOATTYPE	**Pbo,
		      FLOATTYPE	**ppto,
		      FLOATTYPE	**Pppto,
		      int	**yid)
{
  int size = 0;

  size += n * 2 * sizeof(FLOATTYPE); /* yo */
  size += n * sizeof(FLOATTYPE); /* ao */
  size += n * sizeof(FLOATTYPE); /* Pao */
  size += sizeof(FLOATTYPE); /* bo */
  size += sizeof(FLOATTYPE); /* Pbo */
  size += 2 * sizeof(FLOATTYPE); /* ppto */
  size += 2 * sizeof(FLOATTYPE); /* Pppto */
  size += n * sizeof(int); /* yid */

  if(!(*yo = (FLOATTYPE *) malloc(size)))
    se_error("Allocation Error");
  *ao = *yo + 2*n;
  *Pao = *ao + n;
  *bo = *Pao + n;
  *Pbo = *bo + 1;
  *ppto = *Pbo + 1;
  *Pppto = *ppto + 2;
  *yid = (int *) (*Pppto + 2);

     /*********************************/
  { /* Fill Required Init Data Lists */
    FLOATTYPE *flp = *yo;
    int *ip = *yid;
    new_type *np;
    int count = 0;

    while(EOF != (int)(np = Fts->get_new(Fts)) && count < n ) {
      *ip++ = np->id;
      *flp++ = np->x;
      *flp++ = np->y;
      count++;
    }
  }

/*** Change ***/ /* So now, .sta files have to match the points
  		    actually used in the estimation, not the points in the
		    .fts file */
  /* Hopefully when I'm done, .sta files do not have to match 
     anything. */

     /*********************************/
  { /* Fill Optional Init Data Lists */
    int count;
    char filename[128];
    FLOATTYPE *alpha_p = *ao;
    FLOATTYPE *varA_p = *Pao;
    depth_type *dp;
    sta_type *Sta;

    for( count=0; count<n; count++ ) {
      *alpha_p++ = SMKF_DEFAULT_A_O;
      *varA_p++ = SMKF_DEFAULT_VAR_A_O;
    }

    sprintf(filename, "%s.sta", params.name);
    Sta = sta_open_read(filename);

    if(NULL != Sta) {
      Sta->scan(Sta);
      params.seq_iter = Sta->iter + 1;

#if 0
      if(Sta->N != n)
	se_error("Sta and Fts first frame npoints must correspond");
#endif /* This should no longer be an error: simply find the
	  feature to which each sta record corresponds and
	  put in the value.  If not found, put in default value */

      while(EOF != (int)(dp = Sta->get_depth(Sta))) {
	int id = dp -> id;
	int *ip = *yid;

	for( count=0; count<n && id!= *ip++; count++ );
	if( count < n ) { /* then we have an fts feature with this id */
	  (*ao)[count] = dp->a;
	  (*Pao)[count] = dp->Va;
	}
	/* else do nothing */
      }
      **bo = Sta->b;
      **Pbo = Sta->Vb;
      (*ppto)[0] = Sta->xp;
      (*ppto)[1] = Sta->yp;
      (*Pppto)[0] = Sta->Vxp;
      (*Pppto)[1] = Sta->Vyp;
      Sta->close(Sta);
    }

    else
      *ao = *Pao = *bo = *Pbo = *ppto = *Pppto = NULL;
    if(params.b >= 0.)
      *bo = &params.b;
    if(params.Pb >= 0.)
      *Pbo = &params.Pb;
  }
}

void
  se_get_known_motion( char *est_file, smkf_state_type *T, FLOATTYPE *Q )
{
  static est_type *Est = NULL;

  if( !Est )
    Est = est_open_read( est_file );

  Est -> scan( Est );
  T->x = Est->x;
  T->y = Est->y;
  T->zb = Est->zb;
  Q[0] = Est->q0;
  Q[1] = Est->q1;
  Q[2] = Est->q2;
  Q[3] = Est->q3;
}

main(int argc, char **argv)
{
  register smkf_type		*SMKF = NULL;
  register fts_type		*Fts;
  register sta_type		*Sta;
  FLOATTYPE			*Yo_pics_p, *Alpha_pics_p, *VarA_pics2_p;
  FLOATTYPE			*Beta_pics_p, *VarB_pics2_p;
  FLOATTYPE			*Ppt_pics_p, *VarPpt_pics2_p;
  int				*Y_id;
  int				N_feat = 0;
  smkf_state_type		known_motion;

  ParseCmdLine(argc, argv);

  Fts = fts_open_read(NULL);				/* Open Feature File */
  Fts->scan(Fts);					/* Read Feature File */

  /* If Nfe < 0 (not valid) use N_feat = N_fts (Fts->N_new)
     else, Nfe is the maximum used, thus N_feat = Min(Nfe, N_fts)

     In either case, se_get_initial_data() should, for each
     of the N_feat features, use initial data from .sta file
     if .sta exists and has a record for the feature id.  If
     not, the feature should be initialized using defaults.

     Thus N_feat is completely determined by Nfe flag and N_fts;
     the .sta file has nothing to do with that.  Thus the .sta file
     can have as little info as it wants.

     Note: if N_feat < N_fts, the first N_feat features from 
     the fts file will be used (.sta file has no effect).
   */

  if( params.nfe > 0 )
    N_feat = (params.nfe < Fts->N_new) ? params.nfe : Fts->N_new;
  else
    N_feat = Fts->N_new;

#if 0
  fprintf( stderr, "max= %d, Fts= %d --> N_feat= %d\n",
	  params.nfe, Fts->N_new, N_feat );
#endif

  se_get_initial_data(Fts, N_feat,			/* Sta File or Dflt  */
		      &Yo_pics_p,
		      &Alpha_pics_p, &VarA_pics2_p,
		      &Beta_pics_p, &VarB_pics2_p,
		      &Ppt_pics_p, &VarPpt_pics2_p,
		      &Y_id);

  SMKF = smkf_make(N_feat,				/* Allocate EKF     */
		   params.name,
		   (params.motionfile) ? &known_motion : NULL);

  if( params.motionfile )
    se_get_known_motion( params.motionfile, &known_motion, SMKF->qp );

  SMKF->initialize(SMKF,				/* Init EKF	     */
		   N_feat,
		   Y_id, Yo_pics_p,
		   Beta_pics_p, VarB_pics2_p,
		   Ppt_pics_p, VarPpt_pics2_p,
		   Alpha_pics_p, VarA_pics2_p,
		   &params.Qr,
		   &params.Qth,
		   &params.Ro);

  SMKF->update(SMKF);					/* Crank EKF 1 step */
  SMKF->print_state(SMKF, stdout);
  if( params.covariance )
    smkf_write_covariance( SMKF, params.name, 0 );

  {			/* MAIN LOOP */
    int Frame = 0;
    while(((params.nfr > 0) ? Frame < params.nfr : 1) 
	  && EOF != Fts->scan(Fts)) {
      Frame++;

      {
	old_type *op;
	register int count;
	
	while(EOF != (int)(op = Fts->get_old(Fts))) {
	  count = N_feat;
	  while( count-- && Y_id[count] != op->id );
	  if( count >= 0 )
	    smkf_new_measurement(SMKF, op->id, &op->x, Frame);
	}
      }
      
      if( params.motionfile )
	se_get_known_motion( params.motionfile, SMKF->motion, SMKF->qp );
      SMKF->update( SMKF );
      SMKF->print_state( SMKF, stdout );
      if( params.covariance )
	smkf_write_covariance( SMKF, params.name, Frame );
    }
    se_print_sta( SMKF );
  }
  exit(0);
}
