/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  smkf.c
  9 December 1992

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    smkf_error()
    smkf_size_in_bytes()
    smkf_malloc()
    smkf_fill_given_memory()
    smkf_make()			; Callable: make data structure 
    smkf_initialize()
    smkf_ekfXloc()
    smkf_new_measurement()
    smkf_update()
    smkf_print_state()
    smkf_print_variance()
    smkf_write_covariance()
    my_matrix_diagonal()
    my_quat_rotation_matrix()
    my_quat_multiply()
    my_vector_dot()
    my_vector_normalize()
    smkf_filter_init()		; Callable: init KF with info from params 
    smkf_linearize()		; Callable: make H given a valid X_p 
    smkf_predict_Y()		; Callable 
    smkf_estimate_X()		; Callable 
    smkf_predict_X()		; Callable 
    smkf_get_predict_X()	; Callable 

  DESCRIPTION:
    Structure/motion Kalman filter.
    Actually also does camera.
    State is motion, point structure and camera.
    Measurements are feature tracks.

  REFERENCES:
    Azarbayejani, Ali and Alex Pentland, "Recursive Estimation of Motion,
       Structure, and Focal Length", IEEE PAMI, June 1995

    Brown, Robert G., << Introduction to Random Signal Analysis and Kalman
        Filtering >>, John Wiley & Sons, New York, 1983

    Gelb, Arthur (Editor), << Applied Optimal Estimation >>, MIT Press,
       Cambridge, MA, 1974

    Friedmann, Martin, Alex Pentland, and Thad Starner, "Device
       Synchronization using an Optimal Linear Filter", Proceedings of the
       ACM Symposium on Interactive 3D Graphics, Cambridge, MA, Vol 26, No
       2, March 1992

*/

#include <stdio.h>
#include <stdlib.h>

#include "matrix/matrix.h"
#include "ekf/ekf.h"
#include "smkf.h"

#define	EXIT_IF_BIG_EULER_ANGLES 1

/****************************************************************\
\****************************************************************/
void
  smkf_error(char *s)
{
  fprintf(stderr, "SMKF_ERROR: %s\n", s);
  exit(0);
}

int
  smkf_size_in_bytes(int n_feat_max)
{
  int size = sizeof(smkf_type);
  
  size += sizeof(smkf_state_type) + n_feat_max*sizeof(FLOATTYPE);  /*xp*/
  size += sizeof(smkf_state_type) + n_feat_max*sizeof(FLOATTYPE);  /*Vxp*/
  size += n_feat_max * sizeof(int);				   /*Xid*/
  size += 2 * n_feat_max * sizeof(FLOATTYPE);			   /*yo*/
  size += n_feat_max * sizeof(int);				   /*Xloc*/
  size += matrix_size_in_bytes(3, 3);				   /*T3x3*/
  size += ekf_size_in_bytes(SMKF_NSMS + n_feat_max, 2*n_feat_max); /*ekf*/

  return size;
}

static smkf_type *
  smkf_malloc(int n_feat_max)
{
  return((smkf_type *) malloc(smkf_size_in_bytes(n_feat_max)));
}

void
  smkf_fill_given_memory
  (smkf_type *f, int n_feat_max, char *name, smkf_state_type *motion )
{
  register int sv_size= sizeof(smkf_state_type) + n_feat_max*sizeof(FLOATTYPE);
  char *cp;

  cp = (char *) (f + 1);

  f->xp = (smkf_state_type *) cp;
  cp += sizeof(smkf_state_type) + n_feat_max*sizeof(FLOATTYPE);

  f->Vxp = (smkf_state_type *) cp;
  cp += sizeof(smkf_state_type) + n_feat_max*sizeof(FLOATTYPE);

  f->Xid = (int *) cp;
  cp += n_feat_max * sizeof(int);

  f->yo = (FLOATTYPE *) cp;
  cp += 2 * n_feat_max * sizeof(FLOATTYPE);

  f->Xloc = (int *) cp;
  cp += n_feat_max * sizeof(int);

  f->T3x3 = (matrix_type *) cp;
  cp += matrix_size_in_bytes(3, 3);

  f->ekf = (ekf_type *) cp;

  f->name = name;
  f->motion = motion;
  f->n_feat_max = n_feat_max;
  f->n_feat = 0;

  fill_matrix_type_given_memory(f->T3x3, 3, 3, "Temp Rotation Matrix");

  ekf_fill_given_memory(f->ekf,
			SMKF_NSMS + n_feat_max,
			2 * n_feat_max,
			smkf_filter_init,
			NULL,
			smkf_linearize,
			smkf_predict_Y,
			smkf_estimate_X,
			NULL,
			( motion ) ? smkf_get_predict_X : smkf_predict_X,
			NULL,
			name);
  
  f->initialize = smkf_initialize;
  f->new_measurement = smkf_new_measurement;
  f->update = smkf_update;
  f->print_state = smkf_print_state;
  f->print_variance = smkf_print_variance;
}
			
smkf_type *
  smkf_make(int n_feat_max, char *name, smkf_state_type *motion)
{
  smkf_type *f = smkf_malloc(n_feat_max);

  smkf_fill_given_memory(f, n_feat_max, name, motion );

  return f;
}

/****************************************************************\
 *
 *   SMKF Functions
 *   
\****************************************************************/
    
void
  smkf_initialize(smkf_type	*f,
#if 0
		  int		w,	/* width of image (Pixels) */
		  int		h,	/* height of image (Pixels) */
#endif
		  int		n,	/* number of features */
		  int		*yid,	/* id's of init feats */
		  FLOATTYPE	*yo,	/* init 2D loc of feats (Pixels) */
	/*Opt*/	  FLOATTYPE	*bo,	/* init Beta (pics) */
	/*Opt*/	  FLOATTYPE	*Pbo,	/* init Beta variance (pics^2) */
	/*Opt*/	  FLOATTYPE	*ppto,	/* init Principal Point (pics) */
	/*Opt*/	  FLOATTYPE	*Pppto,	/* init Ppt variance (pics^2) */
	/*Opt*/	  FLOATTYPE	*ao,	/* init Alphas (pics) */
	/*Opt*/	  FLOATTYPE	*Pao,	/* init Alpha var's (pics^2) */
	/*Opt*/	  FLOATTYPE	*Qro,	/* pics^2 *//* Amt of int.frame Tr */
	/*Opt*/	  FLOATTYPE	*Qtho,	/* pics^2 *//* Amt of int.frame Rot*/
	/*Opt*/	  FLOATTYPE	*Ro)	/* pics^2 *//* Amt of Meas Noise   */
{
  if(NULL == f || n <= 0 || n> f->n_feat_max || NULL == yo)
    smkf_error("smkf_initialize: bad arguments");

#if 0
  f->pixels_per_pic = (w > h) ? w : h;
  f->bo   = (bo)   ? *bo   : SMKF_DEFAULT_B_O;
  f->Pbo  = (Pbo)  ? *Pbo  : SMKF_DEFAULT_VAR_B_O;
#endif
  
  /* if known motion, f->motion must be set to initial motion
     values before this function is called (obviously) */
  SMKF_X(f->xp)   = ( f->motion ) ? SMKF_X(f->motion) : SMKF_DEFAULT_X_O;
  SMKF_Y(f->xp)   = ( f->motion ) ? SMKF_Y(f->motion) : SMKF_DEFAULT_Y_O;
  SMKF_ZB(f->xp)  = ( f->motion ) ? SMKF_ZB(f->motion) : SMKF_DEFAULT_ZB_O;
  SMKF_THX(f->xp) = ( f->motion ) ? 0. : SMKF_DEFAULT_THX_O;
  SMKF_THY(f->xp) = ( f->motion ) ? 0. : SMKF_DEFAULT_THY_O;
  SMKF_THZ(f->xp) = ( f->motion ) ? 0. : SMKF_DEFAULT_THZ_O;
  SMKF_B(f->xp)   = (bo) ? *bo : SMKF_DEFAULT_B_O;
  SMKF_PX(f->xp)  = (ppto) ? ppto[0] : SMKF_DEFAULT_PX_O;
  SMKF_PY(f->xp)  = (ppto) ? ppto[1] : SMKF_DEFAULT_PY_O;

  SMKF_X(f->Vxp)   = ( f->motion ) ? 0. : SMKF_DEFAULT_VAR_X_O;
  SMKF_Y(f->Vxp)   = ( f->motion ) ? 0. : SMKF_DEFAULT_VAR_Y_O;
  SMKF_ZB(f->Vxp)  = ( f->motion ) ? 0. : SMKF_DEFAULT_VAR_ZB_O;
  SMKF_THX(f->Vxp) = ( f->motion ) ? 0. : SMKF_DEFAULT_VAR_THX_O;
  SMKF_THY(f->Vxp) = ( f->motion ) ? 0. : SMKF_DEFAULT_VAR_THY_O;
  SMKF_THZ(f->Vxp) = ( f->motion ) ? 0. : SMKF_DEFAULT_VAR_THZ_O;
  SMKF_B(f->Vxp)   = (Pbo) ? *Pbo : SMKF_DEFAULT_VAR_B_O;
  SMKF_PX(f->Vxp)  = (Pppto) ? Pppto[0] : SMKF_DEFAULT_VAR_PX_O;
  SMKF_PY(f->Vxp)  = (Pppto) ? Pppto[1] : SMKF_DEFAULT_VAR_PY_O;

  /* This fills in the Alpha portions of f->xp and f->Vxp.
     Also fills in f->yo (init feature locations).
     Also fills in f->Xid, so we can id where features are in X.
   */
  f->n_feat = 0;
  while(n-- && f->n_feat < f->n_feat_max) {
    int ekfXloc = f->n_feat++;

    f->Xid[ekfXloc] = *yid++;

    f->yo[2 * ekfXloc] = yo[0];
    f->yo[2 * ekfXloc + 1] = yo[1];
    yo += 2;

    SMKF_ALPHA(f->xp, ekfXloc) = (ao) ? *ao++ : SMKF_DEFAULT_A_O;
    /* This fixes one of the alphas (the first).  Kind of a HACK */
    SMKF_ALPHA(f->Vxp, ekfXloc) = (Pao) ? *Pao++ : 
      (1 == f->n_feat) ? 0. : SMKF_DEFAULT_VAR_A_O;
  }

  f->Qro  = (Qro)  ? *Qro  : SMKF_DEFAULT_VAR_TR_NOISE;
  f->Qtho = (Qtho) ? *Qtho : SMKF_DEFAULT_VAR_ROT_NOISE;
  f->Ro   = (Ro)   ? *Ro   : SMKF_DEFAULT_VAR_MEAS_NOISE;

  f->tp = 0;
  if( !f->motion ) {
    f->qp[0] = 1.; f->qp[1] = 0.; f->qp[2] = 0.; f->qp[3] = 0.; 
  }
  /* If known motion is being read in, this quaternion is set
     externally (uh-oh, tricky...) */

  f->te = (-1);

  f->n_meas = 0;

  f->ekf->initialize(f->ekf, f);
}

static int
  smkf_ekfXloc(smkf_type *f, int smkf_id)
{
  register *ip = f->Xid, count=0;

  while(*ip++ != smkf_id && count < f->n_feat)
    count++;
  if(count == f->n_feat)
    return EOF;
  else
    return count;
}

void
  smkf_new_measurement(smkf_type *f,
		       int smkf_id,
		       FLOATTYPE *y,/* pics */
		       int t)
/* Pushes new meas pair into f->ekf->Y and pushes correct ekfXloc
   into f->Xloc */
{
  register int ekfYloc;
  register int ekfXloc = smkf_ekfXloc(f, smkf_id);
  
  if(smkf_id < 0 || t != f->tp || f->n_meas >= f->n_feat_max || EOF==ekfXloc)
    smkf_error("smkf_new_measurement: we got bugs");

  ekfYloc = f->n_meas++;
  f->Xloc[ekfYloc] = ekfXloc;
  f->ekf->Y[2*ekfYloc] = y[0];
  f->ekf->Y[2*ekfYloc + 1] = y[1];
}

void
  smkf_update(smkf_type *f)
{
  ekf_resize_filter(f->ekf, SMKF_NSMS + f->n_feat, 2*f->n_meas);
  ekf_filter_step(f->ekf, f);
  
  f->tp++;
  f->te++;
  f->n_meas = 0;
}

void
  smkf_print_state(smkf_type *f, FILE *stream)
{
  register int ekfXloc;
  FLOATTYPE	x, y;

  fprintf(stream, "%.10g %.10g %.10g\n",
	  SMKF_X(f->ekf->X_e), SMKF_Y(f->ekf->X_e), SMKF_ZB(f->ekf->X_e));

  fprintf(stream, "%.10g %.10g %.10g %.10g\n",
	  f->qe[0], f->qe[1], f->qe[2], f->qe[3]);

  fprintf(stream, "%.10g\n",
	  SMKF_B(f->ekf->X_e));

  fprintf(stream, "%.10g %.10g\n",
	  SMKF_PX(f->ekf->X_e), SMKF_PY(f->ekf->X_e));

  fprintf(stream, "%d\n", f->n_feat);
  for(ekfXloc = 0; ekfXloc < f->n_feat; ekfXloc++)
    fprintf(stream, "  %d %.10g\n",
	    f->Xid[ ekfXloc ],
	    SMKF_ALPHA(f->ekf->X_e, ekfXloc));
}

void
  smkf_print_variance(smkf_type *f, FILE *stream)
{
  register int ekfXloc;
  FLOATTYPE	x, y;

  fprintf(stream, "%.10g %.10g %.10g\n",
	  SMREF(f->ekf->P_e, 0, 0),
	  SMREF(f->ekf->P_e, 1, 1),
	  SMREF(f->ekf->P_e, 2, 2));

  fprintf(stream, "%.10g %.10g %.10g %.10g\n",
	  0.,
	  SMREF(f->ekf->P_e, 3, 3),
	  SMREF(f->ekf->P_e, 4, 4),
	  SMREF(f->ekf->P_e, 5, 5));
	  
  fprintf(stream, "%.10g\n", SMREF(f->ekf->P_e, 6, 6));

  fprintf(stream, "%.10g %.10g\n",
	  SMREF(f->ekf->P_e, 7, 7),
	  SMREF(f->ekf->P_e, 8, 8));

  fprintf(stream, "%d\n", f->n_feat);
  for(ekfXloc = 0; ekfXloc < f->n_feat; ekfXloc++)
    fprintf(stream, "  %d %.10g\n",
	    f->Xid[ ekfXloc ],
	    SMKF_VAR_ALPHA(f->ekf->P_e, ekfXloc));
}

#if 0
void
  smkf_write_covariance( smkf_type *f, char *fileroot, int frame )
/* This is a "Stan" addition.  Contrary to what might be expected, this
   function ouputs the lower NxN chunk of structure covariance, not the
   whole covariance matrix.  This is because that's all Stan wants. */
{
  sub_matrix_type *P = f->ekf->P_e;
  FLOATTYPE *flp;
  int row = 9;
  FILE *fp;
  char name[128];

  sprintf( name, "%s.covariance/cov.%.3d", fileroot, frame );
  fp = fopen( name, "w" );

  if( fp ) {
    while( row < P->h ) {
      fwrite( &SMREF( P, row, 9 ), sizeof(float), P->w - 9, fp );
      row++;
    }
    fclose( fp );
  }
  else
    smkf_error( "write_covariance: couldn't write file" );
}
#endif

#define ROWSTART 0

void
  smkf_write_covariance( smkf_type *f, char *fileroot, int frame )
/* This is the "Trevor" edition.  Contrary to what might be expected, this
   function ouputs the entire covariance matrix, not just the lower NxN chunk
   of structure covariance.  This is because that. */
/* only difference between this function and above function is
   ROWSTART = 9 in above function. Btw, ROWSTART is a misnomer
   because it starts the row and the column */
{
  sub_matrix_type *P = f->ekf->P_e;
  FLOATTYPE *flp;
  int row = ROWSTART;
  FILE *fp;
  char name[128];
  int i, j;

  sprintf( name, "%s.covariance/cov.%.3d", fileroot, frame );
  fp = fopen( name, "w" );

  if( fp ) {
/* OLD
    while( row < P->h ) {
      fwrite( &SMREF( P, row, ROWSTART ), sizeof(float), P->w - ROWSTART, fp );
      row++;
    }
*/
    for (i=0; i<P->h; i++) {
      for (j=0; j<P->w; j++)
         fprintf(fp, "%f ", (&(SMREF( P, i, ROWSTART)))[j]); 
      fprintf(fp, "\n");
    }
    fclose( fp );
  }
  else
    smkf_error( "write_covariance: couldn't write file" );
}


#if 0
void
  smkf_print_sta_file(smkf_type *f)
{
  char filename[128];
  FILE *fp;
  int count, iter = 0;

  sscanf(f->name, "%s", filename);
  strcat(filename, ".sta");

  fp = fopen(filename, "r");
  if(fp) {
    fscanf(fp, "%*s%d", &iter);
    iter++;
    fclose(fp);
  }

  fp = fopen(filename, "w");
  if(!fp)
    smkf_error("Cannot open .sta file for write");

  fprintf(fp, "ITERATION %d\n", iter);
  fprintf(fp, "BETA %.10g %.10g\n",
	  SMKF_BETA(f->ekf->X_e), SMKF_VAR_BETA(f->ekf->P_e));
  fprintf(fp, "PPT %.10g %.10g\n",
	  SMKF_PX(f->ekf->X_e), SMKF_PY(f->ekf->X_e));
  fprintf(fp, "STRUCTURE_ID_A_VA_Y1_Y2 %d\n", f->n_feat);
  for(count=0; count<f->n_feat; count++) {
    fprintf(fp, "  %d %.10g %.10g %.10g %.10g\n",
	    count,
	    SMKF_ALPHA(f->ekf->X_e, count),
	    SMKF_VAR_ALPHA(f->ekf->P_e, count),
	    f->yo[2*count],
	    f->yo[2*count + 1]);
  }
  fclose(fp);
}

void
  smkf_scan_sta_file(FILE *stream,
	/*Opt*/	     int *iter_p,
	/*Opt*/	     FLOATTYPE *beta_p, FLOATTYPE *varb_p,
	/*Opt*/	     FLOATTYPE *px_p, FLOATTYPE *py_p,
	/*Opt*/	     int *n_p,
	/*Opt*/	     int *ids_p,
	/*Opt*/	     FLOATTYPE *alphas_p,
	/*Opt*/	     FLOATTYPE *vara_p,
	/*Opt*/	     FLOATTYPE *y_p)
{
  int n;

  if(!stream)
    smkf_error("smkf_scan_sta_file: NULL file pointer");

  fscanf(stream, (iter_p) ? "%*s%d" : "%*s%*d" , iter_p);
  fscanf(stream, (beta_p) ? "%*s%g" : "%*s%*g" , beta_p);
  fscanf(stream, (varb_p) ? "%g" : "%*g" , varb_p);
  fscanf(stream, (px_p) ? "%*s%g" : "%*s%*g" , px_p);
  fscanf(stream, (py_p) ? "%g" : "%*g" , py_p);
  fscanf(stream, "%*s%d", &n);
  if(n_p)
    *n_p = n;
  while(n--) {
    if(ids_p) fscanf(stream, "%d", ids_p++); else fscanf(stream, "%*d");
    if(alphas_p) fscanf(stream, "%f", alphas_p++); else fscanf(stream, "%*f");
    if(vara_p) fscanf(stream, "%f", vara_p++); else fscanf(stream, "%*f");
    if(y_p) {
      fscanf(stream, "%f", y_p++);
      fscanf(stream, "%f", y_p++);
    }
    else fscanf(stream, "%*f%*f");
  }
}
#endif

/*****************************************************************\
 *
 *   Auxilliary Functions
 *
\*****************************************************************/
static void
  my_matrix_diagonal(matrix_type *m, FLOATTYPE *v)
{
  register int count;
  register FLOATTYPE *p, *vec_p;

  matrix_zero(m);

  count = m->w;
  p = m->m;
  vec_p = v;
  Duff4(count, *p++ = *vec_p++; p += m->w);
}

static matrix_type *
  my_quat_rotation_matrix(FLOATTYPE *q, matrix_type *m)
{
  FLOATTYPE q01, q02, q03, q11, q12, q13, q22, q23, q33, *p;

  if(!q || !m || 3 != m->w || 3 != m->h)
    return NULL;
  q01 = q[0]*q[1];
  q02 = q[0]*q[2];
  q03 = q[0]*q[3];
  q11 = q[1]*q[1];
  q12 = q[1]*q[2];
  q13 = q[1]*q[3];
  q22 = q[2]*q[2];
  q23 = q[2]*q[3];
  q33 = q[3]*q[3];
  
  p = m->m;
  *p++ = 1.0 - 2 * (q22 + q33);
  *p++ =     - 2 * (q03 - q12);
  *p++ =       2 * (q02 + q13);
  *p++ =       2 * (q03 + q12);
  *p++ = 1.0 - 2 * (q11 + q33);
  *p++ =     - 2 * (q01 - q23);
  *p++ =     - 2 * (q02 - q13);
  *p++ =       2 * (q01 + q23);
  *p++ = 1.0 - 2 * (q11 + q22);

  return m;
}

static void
  my_quat_multiply(FLOATTYPE  *q1, FLOATTYPE  *q2, FLOATTYPE  *q3)
{
  FLOATTYPE              qt[4];

  if(NULL == q3 || q1 == q3 || q2 == q3) {
    qt[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3];
    qt[1] = q1[1]*q2[0] + q1[0]*q2[1] - q1[3]*q2[2] + q1[2]*q2[3];
    qt[2] = q1[2]*q2[0] + q1[3]*q2[1] + q1[0]*q2[2] - q1[1]*q2[3];
    qt[3] = q1[3]*q2[0] - q1[2]*q2[1] + q1[1]*q2[2] + q1[0]*q2[3];
    q3[0] = qt[0];
    q3[1] = qt[1];
    q3[2] = qt[2];
    q3[3] = qt[3];
  }
  else {
    q3[0] = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3];
    q3[1] = q1[1]*q2[0] + q1[0]*q2[1] - q1[3]*q2[2] + q1[2]*q2[3];
    q3[2] = q1[2]*q2[0] + q1[3]*q2[1] + q1[0]*q2[2] - q1[1]*q2[3];
    q3[3] = q1[3]*q2[0] - q1[2]*q2[1] + q1[1]*q2[2] + q1[0]*q2[3];
  }
}

static FLOATTYPE
  my_vector_dot(FLOATTYPE *v1, FLOATTYPE *v2, int n)
{
  register FLOATTYPE sum = 0.;

  while(n--)
    sum += (*v1++) * (*v2++);
  return sum;
}

static void
  my_vector_normalize(FLOATTYPE *invector, FLOATTYPE *outvector, int n )
{
  FLOATTYPE length;

  length = sqrt ( my_vector_dot(invector, invector, n) );

  if ( (length < 0.00000001) && (length > -0.00000001) )
    vector_zero(outvector, n);
  else
    while(n--)
      outvector[n] = invector[n] / length;
}

/*************************************************************************\
 *
 *   EKF functions
 *
\*************************************************************************/

/*************************************************************************\
 * smkf_filter_init: Initializes size of ekf, Xp, Pp, Q, R, and Phi 
\*************************************************************************/
void
  smkf_filter_init(ekf_type *f, smkf_type *p)
{
  ekf_resize_filter(f, SMKF_NSMS + p->n_feat, 0);
  
  vector_copy((FLOATTYPE *)p->xp, f->X_p, f->n);	/* X_p	*/

  my_matrix_diagonal(f->P_p->M, (FLOATTYPE *)p->Vxp);	/* P_p	*/

  sub_matrix_zero(f->Q);				/* Q	*/
  SMREF(f->Q, 0, 0) = p->Qro;
  SMREF(f->Q, 1, 1) = p->Qro;
  SMREF(f->Q, 2, 2) = p->Qro * SMKF_B(p->xp) * SMKF_B(p->xp);
  SMREF(f->Q, 3, 3) = p->Qtho;
  SMREF(f->Q, 4, 4) = p->Qtho;
  SMREF(f->Q, 5, 5) = p->Qtho;

  matrix_identity(f->R->M);				/* R	*/
  matrix_scale(f->R->M, p->Ro, f->R->M);

  matrix_identity(f->Phi->M);				/* Phi	*/
}

/*************************************************************************\
 * smkf_linearize: Computes H(t) from X_p(t) and Structure
\*************************************************************************/
#define Q0 (p->qp[0])
#define Q1 (p->qp[1])
#define Q2 (p->qp[2])
#define Q3 (p->qp[3])

void
  smkf_linearize(ekf_type *f, smkf_type *p)
{
  static   FLOATTYPE	q00, q01, q02, q03, q11, q12, q13, q22, q23, q33,
  			c0, c1, c2, c3, c4, c5, k0, k1, k2,
  			pc1, pc2, pc3B,
  			Rchph1, Rchph2, Rchph3,
  			ph1, ph2, ph3,
  			m0, m1, m2,
  			y1, y2,
  			v1, v2, v3,
  			Rv1, Rv2, Rv3,
  			g1, g2, g3,
  			dTdO[9];
  register int		ekfYloc, ekfXloc, row_width = f->H->line;
  register FLOATTYPE	*alpha_p, *Hx_p, *Hy_p;
  register FLOATTYPE	*pt, alpha, temp, beta;

  if(2*p->n_meas != f->p || SMKF_NSMS + p->n_feat != f->n)
    smkf_error("smkf_linearize");

  matrix_zero(f->H->m);

  q00 = Q0*Q0;  q01 = Q0*Q1;  q02 = Q0*Q2;  q03 = Q0*Q3;
                q11 = Q1*Q1;  q12 = Q1*Q2;  q13 = Q1*Q3;
                              q22 = Q2*Q2;  q23 = Q2*Q3;
                                            q33 = Q3*Q3;

  c0 = 2*(q01 + q23);  c1 = 2*(q01 - q23);
  c2 = 2*(q02 + q13);  c3 = 2*(q02 - q13);
  c4 = 2*(q03 + q12);  c5 = 2*(q03 - q12);

  k0 = q00 + q11 - q22 - q33;
  k1 = q00 - q11 + q22 - q33;
  k2 = q00 - q11 - q22 + q33;

  Hx_p = f->H->m;
  Hy_p = Hx_p + row_width;
  beta = SMKF_B(f->X_p);

  for(ekfYloc=0; ekfYloc<p->n_meas; ekfYloc++) {
    ekfXloc = p->Xloc[ekfYloc];
    alpha = SMKF_ALPHA(f->X_p, ekfXloc);
    
    y1 = p->yo[ 2*ekfXloc ] - SMKF_PX(f->X_p);
    y2 = p->yo[ 2*ekfXloc + 1 ] - SMKF_PY(f->X_p);

    ph1 = (1. + alpha * beta) * y1;
    ph2 = (1. + alpha * beta) * y2;
    ph3 = alpha;

    Rchph1 = k0*ph1 - c5*ph2 + c2*ph3;
    Rchph2 = c4*ph1 + k1*ph2 - c1*ph3;
    Rchph3 = (-c3)*ph1 + c0*ph2 + k2*ph3;
     
    pc1 = Rchph1 + SMKF_X(f->X_p);
    pc2 = Rchph2 + SMKF_Y(f->X_p);
    pc3B = beta*Rchph3 + SMKF_ZB(f->X_p);
     
    m0 = 1. / (1. + pc3B);
    m1 = ( -m0 * m0 * pc1 );
    m2 = ( -m0 * m0 * pc2 );

    SMKF_X(Hx_p) = m0;
    SMKF_ZB(Hx_p) = m1;
    SMKF_Y(Hy_p) = m0;
    SMKF_ZB(Hy_p) = m2;

    pt = dTdO; /* compute d(transform)/d(omega) */
    *pt++ = ph2*c2+ph3*c5;  *pt++ = -ph1*c2+ph3*k0;  *pt++ = -ph1*c5-ph2*k0;
    *pt++ = -ph2*c1-ph3*k1; *pt++ = ph1*c1+ph3*c4;   *pt++ = ph1*k1-ph2*c4;
    *pt++ = beta*( ph2*k2-ph3*c0 );
    *pt++ = beta*( -ph1*k2-ph3*c3 );
    *pt   = beta*( ph1*c0+ph2*c3 );

    /* compute dh/d(omega) */
    SMKF_THX(Hx_p) = m0 * dTdO[0] + m1 * dTdO[6];
    SMKF_THY(Hx_p) = m0 * dTdO[1] + m1 * dTdO[7];
    SMKF_THZ(Hx_p) = m0 * dTdO[2] + m1 * dTdO[8];
    SMKF_THX(Hy_p) = m0 * dTdO[3] + m2 * dTdO[6];
    SMKF_THY(Hy_p) = m0 * dTdO[4] + m2 * dTdO[7];
    SMKF_THZ(Hy_p) = m0 * dTdO[5] + m2 * dTdO[8];

    v1 = alpha*( k0*y1 - c5*y2 );
    v2 = alpha*( c4*y1 + k1*y2 );
    v3 = Rchph3 + alpha*beta*(-c3*y1 + c0*y2);

    SMKF_B(Hx_p) = m0*v1 + m1*v3;
    SMKF_B(Hy_p) = m0*v2 + m2*v3;

    SMKF_PX(Hx_p) = 1.;
    SMKF_PY(Hy_p) = 1.;

    v1 = beta * y1;
    v2 = beta * y2;
    v3 = 1.;
    Rv1 = ( k0*v1 - c5*v2 + c2*v3);
    Rv2 = ( c4*v1 + k1*v2 - c1*v3);
    Rv3 = beta*(-c3*v1 + c0*v2 + k2*v3);
    SMKF_ALPHA(Hx_p, ekfXloc) = m0*Rv1 + m1*Rv3;
    SMKF_ALPHA(Hy_p, ekfXloc) = m0*Rv2 + m2*Rv3;

    Hx_p = Hy_p + row_width;
    Hy_p = Hx_p + row_width;
  }
}

/*************************************************************************\
 * smkf_predict_Y: computes y_p(t) from x_p(t) and structure
\*************************************************************************/
void 
  smkf_predict_Y(ekf_type *f, smkf_type *p)
{
  int                  	ekfYloc, ekfXloc;
  FLOATTYPE		pO[3], pC[3], alpha, beta, *yp_p;
  FLOATTYPE		y1, y2;

  yp_p = f->Y_p;

  for(ekfYloc=0; ekfYloc<p->n_meas; ekfYloc++) {
    ekfXloc = p->Xloc[ekfYloc];

    alpha = SMKF_ALPHA(f->X_p, ekfXloc);
    beta = SMKF_B(f->X_p);
    y1 = p->yo[ 2*ekfXloc ] - SMKF_PX(f->X_p);
    y2 = p->yo[ 2*ekfXloc + 1 ] - SMKF_PY(f->X_p);

    pO[0] = (1. + alpha * beta) * y1;
    pO[1] = (1. + alpha * beta) * y2;
    pO[2] = alpha;

    my_quat_rotation_matrix(p->qp, p->T3x3);
    matrix_multiply_vector(p->T3x3, pO, pC);

    pC[0] += SMKF_X(f->X_p);
    pC[1] += SMKF_Y(f->X_p);
    pC[2] =  beta * pC[2] + SMKF_ZB(f->X_p);

    *yp_p++ = pC[0] / (1. + pC[2]) + SMKF_PX(f->X_p);
    *yp_p++ = pC[1] / (1. + pC[2]) + SMKF_PY(f->X_p);
  }
}

/*************************************************************************\
 * smkf_estimate_X: Computes X_e(t) and manages the quaternion.
\*************************************************************************/
void
  smkf_estimate_X(ekf_type *f, smkf_type *p)
{
  FLOATTYPE q_c[4];

  vector_subtract( f->Y, f->Y_p, f->p, f->dY );
  sub_matrix_multiply_vector( f->K, f->dY, f->dX );
  vector_add( f->X_p, f->dX, f->n, f->X_e );

  /* Compose RPY_e rotation estimate with q_p to get q_e, *\
  \*   then set RPY_e to zero				  */
  q_c[1] = SMKF_THX(f->X_e) / 2.0;
  q_c[2] = SMKF_THY(f->X_e) / 2.0;
  q_c[3] = SMKF_THZ(f->X_e) / 2.0;
  q_c[0] = 1.0 - q_c[1]*q_c[1] - q_c[2]*q_c[2] - q_c[3]*q_c[3];
  if(q_c[0] < 0) {
    fprintf(stderr, "smkf_estimate_X: big Euler angles (%f, %f, %f)\n",
	    SMKF_THX(f->X_e), SMKF_THY(f->X_e), SMKF_THZ(f->X_e));
#if EXIT_IF_BIG_EULER_ANGLES
    exit( 1 );
#endif
  }
  q_c[0] = sqrt(fabs((FLOATTYPE) q_c[0]));
  my_quat_multiply(p->qp, q_c, p->qe);
  my_vector_normalize(p->qe, p->qe, 4);
  vector_zero(((smkf_state_type *)f->X_e)->rpy, 3);
}

/*************************************************************************\
 * smkf_predict_X: Computes X_p(t) and manages the quaternion.
\*************************************************************************/
void
  smkf_predict_X(ekf_type *f, smkf_type *p)
{
  vector_copy(f->X_e, f->X_p, f->n);
  vector_copy(p->qe, p->qp, 4);
  vector_zero(((smkf_state_type *)f->X_p)->rpy, 3);
}

void
  smkf_get_predict_X( ekf_type *f, smkf_type *p )
{
  /* Note: p->motion(x,y,zb) and p->qp(0,1,2,3) must be set
     before calling smkf_update.  All we do is copy these in */

  vector_copy( p->motion, f->X_p, 3 );
  vector_zero(((smkf_state_type *)f->X_p)->rpy, 3);
}
