/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  smkf.h
  9 October 1992

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    typedef smkf_state_type
    typedef smkf_type

  DESCRIPTION:
    Header file for smkf.o: Structure/Motion Kalman Filter.

*/

#include <math.h>
#include "matrix/matrix.h"

#ifndef _SMKF_H_
#define _SMKF_H_

/*
 *   All lengths are expressed in units of "pic"s
 *   1 "pic" = maximum dimension of the image in pixels
 *
 *   When the Focal Length exceeds 100 times the image size,
 *   you have essentially ORTHOGRAPHIC projection.  These
 *   constants allow ORTHOGRAPHIC projections to be detected.
 */

#define	SMKF_F_BIG		(100.)	/* pics */
#define	SMKF_BETA_SMALL	(.01)	/* pics^-1 */
#define	SMKF_BETA_ORTHO(b)	(b < SMKF_BETA_SMALL)

#define	SMKF_DEFAULT_VAR_TR_NOISE	(.1)
#define	SMKF_DEFAULT_VAR_ROT_NOISE	(.01)
#define	SMKF_DEFAULT_VAR_MEAS_NOISE	(.00006)

#define	SMKF_DEFAULT_X_O		(0.)
#define	SMKF_DEFAULT_Y_O		(0.)
#define	SMKF_DEFAULT_ZB_O		(0.)
#define	SMKF_DEFAULT_THX_O		(0.)
#define	SMKF_DEFAULT_THY_O		(0.)
#define	SMKF_DEFAULT_THZ_O		(0.)
#define	SMKF_DEFAULT_B_O		(2.)
#define SMKF_DEFAULT_PX_O		(0.)
#define SMKF_DEFAULT_PY_O		(0.)
#define	SMKF_DEFAULT_A_O		(1.)

#define	SMKF_DEFAULT_VAR_X_O		(0.)
#define	SMKF_DEFAULT_VAR_Y_O		(0.)
#define	SMKF_DEFAULT_VAR_ZB_O		(0.)
#define	SMKF_DEFAULT_VAR_THX_O		(0.)
#define	SMKF_DEFAULT_VAR_THY_O		(0.)
#define	SMKF_DEFAULT_VAR_THZ_O		(0.)
#define	SMKF_DEFAULT_VAR_B_O		(4.)
#define	SMKF_DEFAULT_VAR_PX_O		(0.)
#define	SMKF_DEFAULT_VAR_PY_O		(0.)
#define	SMKF_DEFAULT_VAR_A_O		(.1)


#define NSTATES_MINUS_STRUCTURE 9 /* Number of Motion and Camera params */
#define	SMKF_NSMS NSTATES_MINUS_STRUCTURE

typedef struct {
  FLOATTYPE	x;		/* X-translation		*/
  FLOATTYPE	y;		/* Y-translation		*/
  FLOATTYPE	zb;		/* Z-translation * Beta		*/
  FLOATTYPE	rpy[3];		/* Roll, pitch, yaw rotation	*/
  FLOATTYPE	b;		/* Beta (== 1/f)		*/
  FLOATTYPE	ppt[2];		/* Principal point		*/
} smkf_state_type;
	/* Structure parameters (alphas)
	   go here at the end		*/

#define SMKF_X(v) (((smkf_state_type *)(v))->x)
#define SMKF_Y(v) (((smkf_state_type *)(v))->y)
#define	SMKF_ZB(v) (((smkf_state_type *)(v))->zb)
#define SMKF_THX(v) (((smkf_state_type *)(v))->rpy[0])
#define SMKF_THY(v) (((smkf_state_type *)(v))->rpy[1])
#define SMKF_THZ(v) (((smkf_state_type *)(v))->rpy[2])
#define SMKF_B(v)  (((smkf_state_type *)(v))->b)
#define SMKF_BETA SMKF_B
#define SMKF_PX(v)  (((smkf_state_type *)(v))->ppt[0])
#define SMKF_PY(v)  (((smkf_state_type *)(v))->ppt[1])

#define SMKF_ALPHA(v, i) (*((FLOATTYPE *)(v) + SMKF_NSMS + i))

/* This is a stupid hack */
#define	SMKF_VAR_BETA(matrix)	SMREF(matrix, 6, 6)
#define SMKF_VAR_PX(matrix)	SMREF(matrix, 7, 7)
#define SMKF_VAR_PY(matrix)	SMREF(matrix, 8, 8)
#define	SMKF_VAR_ALPHA(matrix, i) SMREF(matrix, SMKF_NSMS+i, SMKF_NSMS+i)

/*
 *   SMKF Data Structure
 *
 *    <Xid> is a (<n_feat>,1) array id'ing the features that are
 *	parameterized in the <ekf> state vector.  <Xid>[n] contains the
 *	integer id of the feature whose alpha value is n'th in <ekf>->X
 *	and whose initial image loc is n'th in <yo>.
 *
 *    <Xloc> is a (<n_meas>,1) array locating the <ekf>->X location of
 *	each measurement in <ekf>->Y.  <Xloc>[m] is in [0,<n_feat>) and
 *	locates the alpha param in <ekf>->X that corresponds to the
 *	m'th measurement in <ekf>->Y.  Function <new_measurement> fills
 *	in <ekf>->Y and uses <Xid> to fill in <Xloc>.
 *	
 */
typedef struct {
  char *		name;		/* Set at MAKE-TIME	*/
  smkf_state_type *	motion;		/* Set at MAKE-TIME	*/
  int			n_feat_max;	/* Set at MAKE-TIME	*/

  /* Static Parameters */
  smkf_state_type *	xp;		/* Set at INIT-TIME	*/
  smkf_state_type *	Vxp;		/* Set at INIT-TIME	*/
  int			n_feat;		/* Set at INIT-TIME	*/
  int *			Xid;		/* Set at INIT-TIME	*/
  FLOATTYPE *		yo;		/* Set at INIT-TIME	*//* pics  */
  FLOATTYPE		Qro;		/* Set at INIT-TIME	*//* pic^2 */
  FLOATTYPE		Qtho;		/* Set at INIT-TIME	*//* pic^2 */
  FLOATTYPE		Ro;		/* Set at INIT-TIME	*//* pic^2 */

  /* Dynamic Parameters */
  int			tp;		/* Init, Update		*/
  FLOATTYPE		qp[4];		/* Init, Update		*/
  int			te;		/* Init, Update		*/
  FLOATTYPE		qe[4];		/* Update		*/
  int			n_meas;		/* Init,New_Meas,Update	*/
  int *			Xloc;		/* New_Meas		*/
  matrix_type *		T3x3;
  ekf_type *		ekf;

  void		(*initialize)();	/* Init all static params */
  void		(*new_measurement)();	/* Build new meas vector */
  void		(*update)();		/* Crank Kalman Filter */
  void		(*print_state)();	/* Print out current state */
  void		(*print_variance)();	/* Print var of current state */
} smkf_type;

/*
 *   SMKF Functions
 *
 */
extern smkf_type *		smkf_make();

extern void			smkf_initialize();  /* SMKF_TYPE functions */
extern void			smkf_new_measurement();
extern void			smkf_update();
extern void			smkf_print_state();
extern void			smkf_print_variance();
extern void			smkf_write_covariance();
extern void			smkf_print_sta_file();

extern void			smkf_filter_init(); /* EKF_TYPE functions */
extern void			smkf_linearize();
extern void			smkf_predict_Y();
extern void			smkf_estimate_X();
extern void			smkf_predict_X();
extern void			smkf_get_predict_X();

#endif /* _SMKF_ */
