/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  sta.c
  26 May 1993

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    sta_error()
    sta_warning()
    sta_list_init()
    sta_list_expand()
    sta_list_check()
    sta_list_input()
    sta_get_depth()
    sta_put_depth()
    sta_scan()
    sta_print()
    sta_close()
    sta_open_read()
    sta_open_write()

  DESCRIPTION:
    Static parameters: structure and camera.

*/

#include <stdio.h>
#include <stdlib.h>
#include "files.h"

               /********************************\
----------------------------------------------------------------
   Utility functions
----------------------------------------------------------------
               \********************************/

static void
  sta_error(char *s)
{
  fprintf(stderr, "sta ERROR: %s\n", s);
  exit(0);
}

static void
  sta_warning(char *s)
{
  fprintf(stderr, "sta WARNING: %s\n", s);
}

               /********************************\
----------------------------------------------------------------
   List handlers.
   These can be taken out and used in general for maintaining
    lists of things.
----------------------------------------------------------------
               \********************************/

static void *
  sta_list_init(size_t obj_size, int *n_total_p)
{
  void * ptr = malloc(FILES_LIST_INIT * obj_size);

  if(NULL == ptr)
    sta_error("sta_list_init: failed to malloc");

  *n_total_p = FILES_LIST_INIT;
  return ptr;
}

static void *
  sta_list_expand(size_t obj_size, void *list, int n_used, int* n_total_p)
{
  void * ptr;

  if(obj_size < 0 || NULL == list || n_used < 0 || *n_total_p <= 0
     || n_used > *n_total_p)
    sta_error("sta_list_expand: stupid args");

  ptr = malloc( (*n_total_p + FILES_LIST_INC) * obj_size );

  if(NULL == ptr)
    sta_error("sta_list_expand: malloc failed");

  bcopy(list, ptr, n_used * obj_size);
  free(list);

  *n_total_p += FILES_LIST_INC;
  return ptr;
}

/* Outputs valid "list" and "*n_total_p" */
static void *
  sta_list_check(void *list, int n_used, int *n_total_p, size_t obj_size)
{
  if(list && n_used < *n_total_p)
    return list;

  else if(NULL == list && 0 == n_used && 0 == *n_total_p && obj_size > 0)
    return sta_list_init(obj_size, n_total_p);

  else if(list && n_used == *n_total_p && obj_size > 0)
    return sta_list_expand(obj_size, list, n_used, n_total_p);

  else
    sta_error("sta_list_check: uh-oh, error in args?");
}

static void
  sta_list_input(void *in, void **l_p, int *nu_p, int *nt_p, size_t obj_size)
{
  void *ptr;

  *l_p = sta_list_check(*l_p, *nu_p, nt_p, obj_size);
  
  ptr = (char *)(*l_p) + (*nu_p) * obj_size;
  bcopy(in, ptr, obj_size);
  (*nu_p)++;
}

               /********************************\
----------------------------------------------------------------
   External functions.
----------------------------------------------------------------
               \********************************/

static depth_type *
  sta_get_depth(sta_type *S)
{
  if(FILES_OPEN_READ == S->mode) {
    if(S->N > 0)
      return( S->depths + --S->N );
    else
      return ((void *)EOF);
  }
  else
    sta_warning("sta_get_depth: file not open for read");
}

static void
  sta_put_depth(sta_type *S, depth_type *dp)
{
  if(FILES_OPEN_WRITE == S->mode)
    sta_list_input((void *)dp, (void **)&S->depths,
		   &S->N, &S->depths_size, sizeof(depth_type));
  else 
    sta_warning("sta_put_depth: file not open for write");
}

static int
  sta_scan(sta_type *S)
{
  int count, nconv, total = 0;
  depth_type dt;

  if(S->mode != FILES_OPEN_READ) {
    sta_warning("sta_scan: File not open for read");
    return EOF;
  }

  nconv = fscanf(S->fp, "%*s%d", &S->iter);
  if(EOF == nconv)
    return EOF;
  else if(1 != nconv)
    sta_error("sta_scan: Read error iteration number");

  if(2 != fscanf(S->fp, "%*s%g%g", &S->b, &S->Vb))
    sta_error("sta_scan: Read error beta");

  if(4 != fscanf(S->fp, "%*s%g%g%g%g", &S->xp, &S->yp, &S->Vxp, &S->Vyp))
    sta_error("sta_scan: Read error principal point");
  
  if(1 != fscanf(S->fp, "%*s%g", &S->dz))
    sta_error("sta_scan: Read error z-shift");
  
  S->N = 0;
  if(1 != fscanf(S->fp, "%*s%d", &count))
    sta_error("sta_scan: Read error depths");
  total += count;
  while(count--) {
    if(5 != fscanf(S->fp, "%d%g%g%g%g", &dt.id, &dt.a, &dt.Va, &dt.x, &dt.y))
      sta_error("sta_scan: Read error depth");
    sta_list_input((void *)&dt, (void **)&S->depths,
		   &S->N, &S->depths_size, sizeof(depth_type));
  }
  return total;
}

static void
  sta_print(sta_type *S)
{
  depth_type *dp;

  if(NULL != S->fp && FILES_OPEN_WRITE == S->mode && S->N >=0) {

    fprintf(S->fp, "ITERATION %d\n", S->iter);
    fprintf(S->fp, "%-6s %14.10g %14.10g\n", "BETA", S->b, S->Vb);
    fprintf(S->fp, "%-6s %14.10g %14.10g %14.10g %14.10g\n",
	    "PPT", S->xp, S->yp, S->Vxp, S->Vyp);
    fprintf(S->fp, "%-6s %14.10g\n", "ZSHIFT", S->dz);

    fprintf(S->fp, "STRUCTURE_ID_A_VA_Y1_Y2 %d\n", S->N);
    while(S->N) {
      dp = S->depths + --S->N;
      fprintf(S->fp, "  %-4d %14.10g %14.10g %14.10g %14.10g\n",
	      dp->id, dp->a, dp->Va, dp->x, dp->y);
    }
    fprintf(S->fp, "\n");
  }
  else
    sta_error("sta_print: bad args");
}

static void
  sta_close(sta_type *S)
{
  if(FILES_OPEN_READ != S->mode && FILES_OPEN_WRITE != S->mode) {
    sta_warning("sta_close: file aint open");
    return;
  }
  S->filename = NULL;
  fclose(S->fp);
  S->fp = NULL;
  S->mode = FILES_CLOSED;
}

sta_type *
  sta_open_read(char *name)
{
  FILE *fp;
  register sta_type *S = NULL;

  fp = (NULL == name) ? stdin : fopen(name, "r");
#ifdef VX
  if(NULL != fp || NULL == name) {
#else
  if(NULL != fp) {
#endif
    S = (sta_type *) malloc(sizeof(sta_type));
    if(NULL != S) {

      S->N = 0;
      S->depths = NULL;
      S->depths_size = 0;

      S->filename = name;
      S->fp = fp;
      S->mode = FILES_OPEN_READ;

      S->get_depth = sta_get_depth;
      S->put_depth = sta_put_depth;
      S->scan = sta_scan;
      S->print = sta_print;
      S->close = sta_close;
    }
    else
      fclose(fp);
  }
  return S;
}

sta_type *
  sta_open_write(char *name)
{
  FILE *fp;
  register sta_type *S = NULL;

  fp = (NULL == name) ? stdout : fopen(name, "w");
  if(NULL != fp) {
    S = (sta_type *) malloc(sizeof(sta_type));
    if(NULL != S) {

      S->iter = 0;
      S->b = 1.;
      S->Vb = 0.;
      S->xp = 0.;
      S->yp = 0.;
      S->Vxp = 0.;
      S->Vyp = 0.;
      S->dz = 0.;

      S->N = 0;
      S->depths = NULL;
      S->depths_size = 0;

      S->filename = name;
      S->fp = fp;
      S->mode = FILES_OPEN_WRITE;

      S->get_depth = sta_get_depth;
      S->put_depth = sta_put_depth;
      S->scan = sta_scan;
      S->print = sta_print;
      S->close = sta_close;
    }
    else
      fclose(fp);
  }
  return S;
}
