/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  stats.c
  Sun Apr 30 1995

  AUTHOR:
    Ali Azarbayejani

  CONTENTS:
    alert()
    error()
    usage()
    parseCommandLine()
    typedef alpha_t
    readStaFile()
    readEstFile()
    translationStats()
    quaternionStats()
    pointStats()
    main()

  DESCRIPTION:
    Computes the mean and rms translational and rotational error given
    the state estimate (.est) and an .est file with an appropriately
    scaled actual trajectory.

  CHANGES by Gavrila: added pointStats(), uv parameters to readStaFile,
                      switched b_actual and b_junk in main(), added
                      boolean switch ground_truth

   Stats includes now Euclidean error reporting. Actual 3-D point locations
   are computed by a) obtaining initial 3-D locations by using <>.sta (u, v)
                      and a_true and a_est
                   b) transforming points by T and Q by <>.act.est and
                      <>.ekf.est
   Assumption is that <>.sta file represents initial 3D location in
   true beta representation (its uv are equal to first entry of <>.fts)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
/* NEW */
#include "matrix/matrix.h"
#include "files.h"
#define sqr(x) ((x) * (x))
#define abs(x) (((x) > 0) ? (x) : (-(x)))

static char *progname = 0;
static char *flags = "[-v]";
static char *help  = "";
static char *io = "< stream > stream";

static int verbose = 0;
static char *actualStaFile = 0;
static char *actualEstFile = 0;
static char *pointStatFile = 0;
static char *estEstFile = 0;
static char *ftsFile = 0;

int ground_truth = 1;    /* NEW, switch on this */


/***************************************************************************/

static void
alert( char *s )
{
  fprintf( stderr, "%s: %s\n", progname, s );
}

static void
error( char *s )
{
  alert( s );
  exit( 1 );
}

usage( int argc, char **argv )
{
  fprintf( stderr, "usage: %s %s\n\t%s\n%s", argv[0], flags, io, help );
  exit( 1 );
}

static void
parseCommandLine( int argc, char **argv )
{
  int i;
  int fileCount = 0;

  progname = argv[0];

  for(i=1; i<argc; i++) 
    {
      if( !strcmp(argv[i], "-v") )
	verbose = 1;
      else if( fileCount != 5 )
	{
	  switch( fileCount )
	    {
	     case 0: actualStaFile = argv[i]; break;
	     case 1: actualEstFile = argv[i]; break;
	     case 2: estEstFile = argv[i]; break;
             case 3: ftsFile = argv[i]; break;
             case 4: pointStatFile = argv[i]; break;
	    }
	  fileCount++;
	}
      else
	usage( argc, argv );
    }
}

/***************************************************************************/

typedef struct
{
  int id;
  float value;
} alpha_t;

/* NEW  added u-v parameters */
static int
readStaFile( char *filename, alpha_t **aArrayP, float *bP, float *u, float *v)
{
  int i;
  int N = 0;
  FILE *stream = fopen( filename, "r" );

  if( !stream )
    error( "Can't open Sta file" );

  fscanf( stream, "%*s%*d" );
  if( bP )
    fscanf( stream, "%*s%f%*f", bP );
  else
    fscanf( stream, "%*s%*f%*f" );
  fscanf( stream, "%*s%*f%*f%*f%*f" );
  fscanf( stream, "%*s%*f" );
  fscanf( stream, "%*s%d", &N );

  if( aArrayP )
    {
      if( !*aArrayP )
	*aArrayP = (alpha_t *) malloc( N * sizeof(alpha_t) );
      if( !*aArrayP )
	error( "readStaFile: failed to allocate alpha array" );
      for( i = 0; i < N; i++ )
	fscanf( stream, "%d%f%*f%f%f", 
                &(*aArrayP)[i].id, &(*aArrayP)[i].value, &(u[i]), &(v[i]));
    }
  else
    for( i = 0; i < N; i++ )
      fscanf( stream, "%*d%*f%*f%*f%*f" );

  fclose( stream );
  return N;
}

static int
readEstFile( char *filename, 
	     float** tP,
	     float** qP,
	     float** bP,
	     alpha_t** aP,
	     int* nPoints )
{
  int frameCount = 0;
  int pointCount = 0;
  int frame;
  int point;
  int N;
  alpha_t* alist;
  FILE* stream = fopen( filename, "r" );

  if( !stream )
    error( "readEstFile: couldn't open est file for input" );

  while( EOF != fscanf( stream, "%*f%*f%*f%*f%*f%*f%*f%*f%*f%*f" ) )
    {
      fscanf( stream, "%d", &N );
      if( frameCount && N != pointCount )
	error( "readEstFile: variable number of features, can't do that to me" );
      pointCount = N;
      while( N-- )
	fscanf( stream, "%*d%*f" );
      frameCount++;
    }
  fclose( stream );

  *tP = ( float * ) malloc( frameCount * 3 * sizeof( float ) );
  *qP = ( float * ) malloc( frameCount * 4 * sizeof( float ) );
  *bP = ( float * ) malloc( frameCount * sizeof( float ) );
  *aP = ( alpha_t * ) malloc( frameCount * pointCount * sizeof( alpha_t ) );

  if( !tP || !qP || !bP || !aP )
    error( "readEstFile: allocation failed" );

  stream = fopen( filename, "r" );

  if( !stream )
    error( "readEstFile: couldn't open est file for input (2)" );

  for( frame = 0; frame < frameCount; frame++ )
    {
      fscanf( stream, 
	      "%f%f%f%f%f%f%f%f%*f%*f",
	      *tP + 3*frame,
	      *tP + 3*frame + 1,
	      *tP + 3*frame + 2,
	      *qP + 4*frame,
	      *qP + 4*frame + 1,
	      *qP + 4*frame + 2,
	      *qP + 4*frame + 3,
	      *bP + frame );
      fscanf( stream, "%*d" ); /* already know it better be pointCount */
      alist = *aP + frame * pointCount;
      for( point = 0; point < pointCount; point++ )
	fscanf( stream, "%d%f", &alist[point].id, &alist[point].value );
    }

  fclose( stream );

  *nPoints = pointCount;
  return frameCount;
}

/* n.b. we assume that pointCount can only decrease from initial frame */
static int
    readFtsFile (char *ftsFile, float *u_measured, float *v_measured)
{
fts_type *Fts, *F;
int Frame = 0;
int pointCount;
old_type *op;
new_type *np;
int i;
char debug_file[81];
int debug = 0;

if (debug)
   printf("Entering readFtsFile\n");

Fts = fts_open_read(ftsFile);
Fts->scan(Fts);

pointCount = Fts->N_new; 
for (i=0; i<pointCount; i++)  {
    np = Fts->get_new(Fts);
    if (np->id >= pointCount) {
       printf("PANIC. point id exceeds pointCount.\n");
       exit(-1);
    }
    u_measured[0*pointCount+pointCount-(np->id+1)] =  np->x;
    v_measured[0*pointCount+pointCount-(np->id+1)] =  np->y;
    if (debug)
       printf("(%6.3f  %6.3f)\n", np->x, np->y);
}

sprintf(debug_file, "debug_%d.txt", Frame);

while (Fts->scan(Fts) != EOF) {
   printf("t = %d\n", Frame);

   Frame++;

   for (i=0; i<pointCount; i++) {
      u_measured[Frame*pointCount+i] = -1;
      v_measured[Frame*pointCount+i] = -1;
   }

   while(EOF != (int)(op = Fts->get_old(Fts))) {
     if (op->id >= pointCount) {
        printf("PANIC. point id exceeds pointCount.\n");
        exit(-1);
     }
     u_measured[Frame*pointCount+pointCount-(op->id+1)] =  op->x;
     v_measured[Frame*pointCount+pointCount-(op->id+1)] =  op->y;
     if (debug)
       printf("(%6.3f  %6.3f)\n", op->x, op->y);
   }
}

return pointCount;
}
         

static void
translationStats( float* truth, float* data, int T, FILE* stream )
{
  float sum = 0.;
  float *tp = truth;
  float *dp = data;
  int count = T * 3;
  float rms;
  float diff;

  while( count-- )
    sum += (*tp++) - (*dp++);

  fprintf( stream, "%-40s:%.10f\n", "Translation: Mean Error", sum / (T*3) );

  sum = 0.;
  tp = truth;
  dp = data;
  count = T * 3;
  while( count-- )
    {
      diff = (*tp++) - (*dp++);
      sum += ( diff * diff );
    }
  sum /= ( T * 3 );
  rms = sqrt( (double) sum );

  fprintf( stream, "%-40s:%.10f\n", "Translation: Mean Square Error", sum );
  fprintf( stream, "%-40s:%.10f\n", "Translation: Root Mean Square Error", rms );
}
  
static void
quaternionStats( float* truth, float* data, int T, FILE* stream )
{
  float sum = 0.;
  float *tp = truth;
  float *dp = data;
  int count = T;
  float rms;
  float angle_diff;

  while( count-- )
    {
      angle_diff = 2 * ( acos((double)*tp) - acos((double)*dp) );
#if 0
      fprintf( stderr, "qa = %f, %f, %f, %f\n", tp[0], tp[1], tp[2], tp[3] );
      fprintf( stderr, "qe = %f, %f, %f, %f\n", dp[0], dp[1], dp[2], dp[3] );
      fprintf( stderr, "da = %f\n", angle_diff );
#endif
      sum += angle_diff;
      tp += 4;
      dp += 4;
    }
  sum /= T;

  fprintf( stream, "%-40s:%.10f rad\n", "Rotation: Mean Error", sum );

  sum = 0.;
  tp = truth;
  dp = data;
  count = T;
  while( count-- )
    {
      angle_diff = 2 * ( acos((double)*tp) - acos((double)*dp) );
      sum += ( angle_diff * angle_diff );
      tp += 4;
      dp += 4;
    }
  sum /= T;
  rms = sqrt( (double) sum );

  fprintf( stream, "%-40s:%.10f rad^2\n", "Rotation: Mean Square Error", sum );
  fprintf( stream, "%-40s:%.10f rad\n", "Rotation: Root Mean Square Error", rms );
}


/* BEGIN NEW */
/* next function copied from synth.c */
static matrix_type *
  my_quat_rotation_matrix(FLOATTYPE *q, matrix_type *m)
{
  FLOATTYPE q01, q02, q03, q11, q12, q13, q22, q23, q33, *p;
/*
  printf("entering my_quat_rot with q = (%6.3f, %6.3f, %6.3f, %6.3f)\n",
         q[0], q[1], q[2], q[3]);
*/
  if(!q || !m || 3 != m->w || 3 != m->h)
    return NULL;
  q01 = q[0]*q[1];
  q02 = q[0]*q[2];
  q03 = q[0]*q[3];
  q11 = q[1]*q[1];
  q12 = q[1]*q[2];
  q13 = q[1]*q[3];
  q22 = q[2]*q[2];
  q23 = q[2]*q[3];
  q33 = q[3]*q[3];

  p = m->m;
  *p++ = 1.0 - 2 * (q22 + q33);
  *p++ =     - 2 * (q03 - q12);
  *p++ =       2 * (q02 + q13);
  *p++ =       2 * (q03 + q12);
  *p++ = 1.0 - 2 * (q11 + q33);
  *p++ =     - 2 * (q01 - q23);
  *p++ =     - 2 * (q02 - q13);
  *p++ =       2 * (q01 + q23);
  *p++ = 1.0 - 2 * (q11 + q22);

  return m;
}

pointStats(float* t_actual, float* t_est, 
           float* q_actual, float* q_est,
           alpha_t* a_actual, alpha_t* a_est,
           float b_actual, float b_sta, float* b_est,
           float* u_sta, float* u_measured,
           float* v_sta, float* v_measured,
           int Nt, int Np, char *filename) 
{
int t, i;
float p_actual[3], p_measured[3];  /* after transformation */ 
float p_act[3], p_est[3], Rp[3];  /* before and after rotation */
float *Q_act, *Q_est, *T_act, *T_est;
float u_act, v_act, u_est, v_est;
matrix_type *R_act, *R_est;
int debug = 0;
FILE *fp;
float errorXYZ, errorXY, errorZ, errorUV, errorUV2;
/* error UV is between true and estimated image coordinates */
/* error UV2 is between measured and estimated image coordinates */

if (ground_truth && abs(b_sta - b_actual) > 0.001) {
   printf("PANIC in pointStats: b_sta = %f and b_actual = %f\n",
           b_sta, b_actual);
   exit(-1);
}

if ((fp = fopen(filename, "w")) == NULL) {
   printf("Error: cannot open %s for writing.\n", filename);
   exit(-1);
}

if (ground_truth) {
   R_act = make_matrix_type( 3, 3, "Rotation Matrix" );
}
   R_est = make_matrix_type( 3, 3, "Rotation Matrix" );

for (t=0; t<Nt; t++) {

   if (ground_truth) {
      Q_act = &(q_actual[t*4]);
      T_act = &(t_actual[t*3]);
      my_quat_rotation_matrix(Q_act, R_act);
   }

   Q_est = &(q_est[t*4]);
   T_est = &(t_est[t*3]);
   my_quat_rotation_matrix(Q_est, R_est);

   if (debug) { 
   fprintf(fp, "t = %d\n", t);

     if (ground_truth) {
       fprintf(fp, "T_act = (%6.3f %6.3f %6.3f)  Q_act = (%6.3f %6.3f %6.3f %6.3f\n",
          T_act[0], T_act[1], T_act[2]/b_actual, 
          Q_act[0], Q_act[1], Q_act[2], Q_act[3]);
     }
     fprintf(fp, "T_est = (%6.3f %6.3f %6.3f)  Q_es = (%6.3f %6.3f %6.3f %6.3f\n",
          T_est[0], T_est[1], T_est[2]/b_est[t], 
          Q_est[0], Q_est[1], Q_est[2], Q_est[3]);
     fprintf(fp, "b_sta = %6.3f, b_actual = %6.3f, b_est = %6.3f\n", 
             b_sta, b_actual, b_est[t]);
     
   } /* end if debug */

  for (i=0; i<Np; i++) {
 
   if (ground_truth) {
     p_act[0] = (1. + a_actual[i].value * b_sta) * u_sta[i];
     p_act[1] = (1. + a_actual[i].value * b_sta) * v_sta[i];
     p_act[2] = a_actual[i].value;
     matrix_multiply_vector(R_act, p_act, Rp);
     Rp[2] *= b_actual;
     vector_add3(Rp, T_act, p_actual);
     p_actual[2] /= b_actual;
     u_act = p_actual[0] / (1. + b_actual*p_actual[2]);
     v_act = p_actual[1] / (1. + b_actual*p_actual[2]);
   }

     p_est[0] = (1. + a_est[t*Np+i].value * b_est[t]) * u_measured[0*Np+i];
     p_est[1] = (1. + a_est[t*Np+i].value * b_est[t]) * v_measured[0*Np+i];
     p_est[2] = a_est[t*Np+i].value;
     matrix_multiply_vector(R_est, p_est, Rp);
     Rp[2] *= b_est[t];
     vector_add3(Rp, T_est, p_measured);
     p_measured[2] /= b_est[t];
     u_est = p_measured[0] / (1. + b_est[t]*p_measured[2]);
     v_est = p_measured[1] / (1. + b_est[t]*p_measured[2]);

     if (debug) {
       fprintf(fp, "i = %d\n", i);

       if (ground_truth) {
        fprintf(fp, "u_sta = %6.3f, v_sta = %6.3f, a_sta %6.3f\n", 
                u_sta[i], v_sta[i], a_actual[i].value);
        fprintf(fp, "P_sta = (%6.3f, %6.3f, %6.3f)\n",
               p_actual[0], p_actual[1], p_actual[2]);
        fprintf(fp, "u_act = %6.3f, v_act = %6.3f\n", u_act, v_act);
       }

       fprintf(fp, "u_measured = %6.3f, v_measured = %6.3f\n",
               u_measured[t*Np+i], v_measured[t*Np+i]);

       fprintf(fp, "a_est %6.3f\n", a_est[t*Np+i].value);
       fprintf(fp, "P_est = (%6.3f, %6.3f, %6.3f)\n",
              p_measured[0], p_measured[1], p_measured[2]);
       fprintf(fp, "u_est = %6.3f, v_est = %6.3f\n", u_est, v_est);
     }

     if (ground_truth) {

      errorXYZ = (float) sqrt((double) (
            sqr(p_actual[0]-p_measured[0]) + 
            sqr(p_actual[1]-p_measured[1]) + 
            sqr(p_actual[2]-p_measured[2])));
      errorXY = (float) sqrt((double) (
            sqr(p_actual[0]-p_measured[0]) + 
            sqr(p_actual[1]-p_measured[1])));
      errorZ = abs((p_actual[2]-p_measured[2]));
      errorUV = (float) sqrt((double) (
            sqr(u_act - u_est) + sqr(v_act - v_est)));
     }
     else 
      errorXYZ = errorXY = errorZ = errorUV = -1.;

     if (u_measured[t*Np+i] > -0.9999 && v_measured[t*Np+i] -0.9999)
        errorUV2 = (float) sqrt((double) (
            sqr(u_measured[t*Np+i] - u_est) + 
            sqr(v_measured[t*Np+i] - v_est)));
     else
        errorUV2 = -1.;
  
     fprintf(fp, "%7.4f  %7.4f  %7.4f  %7.5f  %7.5f\n", 
           errorXYZ, errorXY, errorZ, errorUV, errorUV2); 
  }
  if (debug)
     fprintf(fp, "\n");
}
fclose(fp);
}
/* END NEW */


main( int argc, char **argv )
{
  int actualPointCount = 0;
  int measuredPointCount = 0;
  int actualFrameCount = 0;
  int estFrameCount = 0;
  int estPointCount = 0;
  int junkCount;

  float *t_actual = NULL; /* T, 3 */
  float *q_actual = NULL; /* T, 4 */
  float b_sta;
  alpha_t *a_junk = NULL;

  float *b_actual = NULL; /* N */
  alpha_t *a_actual = NULL; /* N */

  float u_actual[1000], v_actual[1000];
  float u_measured[50000], v_measured[50000];
  
  float *t_est = NULL; /* T, 3 */
  float *q_est = NULL; /* T, 4 */
  float *b_est = NULL; /* T, 1 */
  alpha_t *a_est = NULL; /* T, N */

  parseCommandLine( argc, argv );

  if( verbose )
    {
      fprintf( stderr, "actualStaFile= %s\n", actualStaFile );
      fprintf( stderr, "actualEstFile= %s\n", actualEstFile );
      fprintf( stderr, "pointStatFile = %s\n", pointStatFile );
      fprintf( stderr, "estEstFile= %s\n", estEstFile );
    }
  
  if (strcmp(actualStaFile, "NULL") == 0) 
     ground_truth = 0;

  fprintf(stderr, "ground_truth = %d\n", ground_truth);

  /* NEW: added uv parameters, switched b_junk and b_actual */
  if (ground_truth)
     actualPointCount = readStaFile( actualStaFile, &a_actual, &b_sta, 
                                     u_actual, v_actual);


  if (ground_truth)
     actualFrameCount = 
       readEstFile( actualEstFile, 
                   &t_actual, &q_actual, &b_actual, &a_junk, &junkCount );

  /* NEW */ 
  measuredPointCount = readFtsFile( ftsFile, u_measured, v_measured);

  estFrameCount = 
    readEstFile( estEstFile, &t_est, &q_est, &b_est, &a_est, &estPointCount );

  if (verbose && ground_truth) {
    int i;

    fprintf( stderr, "beta= %f\n", b_actual[0]);

    for( i= 0; i<actualPointCount; i++ )
      fprintf( stderr, "a[%d]= %d, %f\n", i, a_actual[i].id, a_actual[i].value
);
  }


  if (estPointCount != actualPointCount && ground_truth)
    error ("estPointCount != actualPointCount");

  if (estPointCount != measuredPointCount)
    error ("estPointCount != measuredPointCount");

  if( verbose )
    {
      fprintf( stderr, "Ta= %d\n", actualFrameCount );
      fprintf( stderr, "Na= %d\n", actualPointCount );
      fprintf( stderr, "Te= %d\n", estFrameCount );
      fprintf( stderr, "Ne= %d\n", estPointCount );
    }

#if 0
  {
    int count = estFrameCount;
    float *flp = b_est;
    while( count-- )
      fprintf( stdout, "%g\n", *flp++ );
  }
  exit(0);
#endif

  fprintf( stdout, "(One unit of distance corresponds to the size of the image)\n" );

  if (ground_truth) {
     translationStats( t_actual, t_est, estFrameCount, stdout);
     quaternionStats( q_actual, q_est, estFrameCount, stdout);
  }
  pointStats(t_actual, t_est, q_actual, q_est,
             a_actual, a_est, b_actual[0], b_sta, b_est,
             u_actual, u_measured, v_actual, v_measured,
             estFrameCount, estPointCount, pointStatFile);

#if 0
  betaStats( b_actual, b_est, estFrameCount, stdout );
  alphaStats( a_actual, a_est, stdout );
#endif
  exit(0);
}

