/*
###########################################################################
# 
# Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
# rights reserved.
# 
# Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
# at the Media Laboratory, MIT, Cambridge, Massachusetts.
# 
# Permission to use, copy,  or modify this software and  its documentation
# for  educational  and  research purposes only and without fee  is hereby
# granted, provided  that this copyright notice and the original authors's
# names appear  on all copies and supporting documentation.  If individual
# files are  separated from  this  distribution directory  structure, this
# copyright notice must be included.  For any other uses of this software,
# in original or  modified form, including but not limited to distribution
# in whole or in  part, specific  prior permission  must be  obtained from
# MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
#   basis  of  a commercial  software  or  hardware product  without first
# obtaining appropriate licenses  from MIT.  MIT. makes no representations
# about the suitability of this  software for any purpose.  It is provided
# "as is" without express or implied warranty.
# 
###########################################################################

  synth.c
  2 June 1993

  AUTHOR:
    Ali J. Azarbayejani

  CONTENTS:
    ParseCmdLine()
    my_quat_rotation_matrix()
    point_from_depth()
    add_random_noise()
    compute_biases()
    compute_features()
    irand48()
    main()

  DESCRIPTION:
    (.sta, .est) --> (.fts)

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "matrix/matrix.h"
#include "files.h"

static matrix_type *RotMat;

static char *usage = "[-v] [-frame #] [-noise <units>] [-bias <units>] [-scale <iscale>]\n\t[-wscale <worldscale>] [-wshift <worldshift>]";

static char *io = "\n\t< (.sta) (.est) > (.fts)";

static char *more = "\n\n\tu = iscale * ((X*wscale)+wshift) / (1 + ((Z*wscale)+wshift) / f)";

static struct {
  int	verbose;
  int	single;
  int	frame;
  char	*fts_name;
  char	*est_name;
  char	*sta_name;
  double noise_level;
  double bias_level;
  double scale;
  double worldscale;
  double worldshift;
  char	*init_name;
  double init_beta;
  double init_Vbeta;	
}
params = { 0, 0, 0, NULL, NULL, NULL, 0., 0., 1., 1., 0., NULL, 1., .25 };

static void
  ParseCmdLine(int argc, char **argv)
{
  int		i;
#if 0
  char		usage[1024], io[128], more[256], *arg;

  sprintf(usage, "%s\n\t%s",
	  "[-v] [-noise <units>] [-bias <units>] [-scale <imagescalar>]",
	  "[-wscale <worldscale>] [-wshift <worldshift>]" );
  sprintf(io,"\n\t%s", "< (.sta) (.est) > (.fts)");
  sprintf(more
#endif


  for(i=1; i<argc; i++) {
    if(!strcmp(argv[i], "-v"))
      params.verbose = 1;
    else if(!strcmp(argv[i], "-frame")) {
      params.frame = atoi(argv[++i]);
      params.single = 1;
    }
    else if(!strcmp(argv[i], "-noise"))
      params.noise_level = atof(argv[++i]);
    else if(!strcmp(argv[i], "-bias"))
      params.bias_level = atof(argv[++i]);
    else if(!strcmp(argv[i], "-scale"))
      params.scale = atof(argv[++i]);
    else if(!strcmp(argv[i], "-wscale"))
      params.worldscale = atof(argv[++i]);
    else if(!strcmp(argv[i], "-wshift"))
      params.worldshift = atof(argv[++i]);
    else if(!strcmp(argv[i], "-init"))
      params.init_name = argv[++i];
    else if(!strcmp(argv[i], "-init_beta"))
      params.init_beta = atof( argv[++i] );
    else if(!strcmp(argv[i], "-init_Vbeta"))
      params.init_Vbeta = atof( argv[++i] );
    else {
      fprintf(stderr,
	      "usage: %s %s %s %s\n",
	      argv[0], usage, io, more);
      exit(1);
    }
  }
}

static matrix_type *
  my_quat_rotation_matrix(FLOATTYPE *q, matrix_type *m)
{
  FLOATTYPE q01, q02, q03, q11, q12, q13, q22, q23, q33, *p;

  if(!q || !m || 3 != m->w || 3 != m->h)
    return NULL;
  q01 = q[0]*q[1];
  q02 = q[0]*q[2];
  q03 = q[0]*q[3];
  q11 = q[1]*q[1];
  q12 = q[1]*q[2];
  q13 = q[1]*q[3];
  q22 = q[2]*q[2];
  q23 = q[2]*q[3];
  q33 = q[3]*q[3];
  
  p = m->m;
  *p++ = 1.0 - 2 * (q22 + q33);
  *p++ =     - 2 * (q03 - q12);
  *p++ =       2 * (q02 + q13);
  *p++ =       2 * (q03 + q12);
  *p++ = 1.0 - 2 * (q11 + q33);
  *p++ =     - 2 * (q01 - q23);
  *p++ =     - 2 * (q02 - q13);
  *p++ =       2 * (q01 + q23);
  *p++ = 1.0 - 2 * (q11 + q22);

  return m;
}

void
  point_from_depth(FLOATTYPE a,
		   FLOATTYPE dx,
		   FLOATTYPE dy, 
		   FLOATTYPE b,
		   FLOATTYPE dz,
		   FLOATTYPE *p)
{
  p[0] = (1. + a * b) * dx * params.worldscale + params.worldshift;
  p[1] = (1. + a * b) * dy * params.worldscale + params.worldshift;
  p[2] = (a + dz )         * params.worldscale + params.worldshift;
}

#define ROUND(x)  ((int) (x + 0.5))

double
  add_random_noise(double input, double noise_level)
{
  extern double drand48();  /*Too bad ain't no header file for drand48*/
  double noise;
  /* drand48 returns (double) in range [0.0, 1.0) */

  noise = drand48(); /* noise in [0., 1.) */
  noise = (noise_level * 2. * (noise - 0.5)); /* noise in [-NL, NL) */
  return ( input + noise );
}

FLOATTYPE *
  compute_biases(sta_type *S, double level)
{
  extern double drand48();  /*Too bad ain't no header file for drand48*/
  int count = 2 * S->N;
  FLOATTYPE *b, *p;

  b = (FLOATTYPE *) malloc(count * sizeof(FLOATTYPE));

  p = b;
  if(b)
    while(count--)
      *p++ = (level * 2. * (drand48() - .5));
  return b;
}

void
  compute_features(sta_type *S, est_type *T, fts_type *F,
		   FLOATTYPE *B, int trash )
/* {S->b,S->xp,S->yp} determine how to get 3D points in object frame.
   {T->b,T->xp,T->yp} determine how to project into image plane
*/
{
  static int	first = 1;
  int 		count = S->N;
  depth_type	*dp, *odp;
  FLOATTYPE	po[3], Rpo[3], pc[3];
  new_type	nt;
  old_type	ot;
  int		min_id, max_id;
  FLOATTYPE	min_a = S->depths->a, max_a = S->depths->a;
  sta_type	*init = NULL;

  if( first && params.init_name )
    init = sta_open_write( params.init_name );

  for( count = 0; count < S->N; count++ ) {
    dp = S->depths + count;

    point_from_depth(dp->a, dp->x - S->xp, dp->y - S->yp, S->b, S->dz, po);
    my_quat_rotation_matrix(&T->q0, RotMat);
    matrix_multiply_vector(RotMat, po, Rpo);
    
    if( first && params.init_name ) {
      if( Rpo[2] < min_a ) { min_a = Rpo[2]; min_id = dp->id; }
      if( Rpo[2] > max_a ) { max_a = Rpo[2]; max_id = dp->id; }
    }

    Rpo[2] *= T->b;
    vector_add3(&T->x, Rpo, pc);

    if(first) {
      nt.id = dp->id;
      nt.x = pc[0] / (1. + pc[2]) * params.scale;
      nt.y = pc[1] / (1. + pc[2]) * params.scale;
      F->put_new(F, &nt);

      if( params.init_name ) {
	init->put_depth( init, dp );
	(init->depths + count)->x = nt.x;
	(init->depths + count)->y = nt.y;
      }
    }
    else {
      ot.id = dp->id;
      ot.x = pc[0] / (1. + pc[2]) * params.scale;
      ot.y = pc[1] / (1. + pc[2]) * params.scale;
      if(params.noise_level > 0.) {
	ot.x = add_random_noise((double)ot.x, params.noise_level);
	ot.x += *B++;
	ot.y = add_random_noise((double)ot.y, params.noise_level);
	ot.y += *B++;
      }
      ot.Vx = pc[2]; /* this is a HACK, for depth cueing */
      ot.Vy = 0;
      F->put_old(F, &ot);
    }
  }
  
  if( first && params.init_name ) {
    init -> iter = 0;
    init -> b = params.init_beta;
    init -> Vb = params.init_Vbeta;
    init -> xp = 0.;
    init -> yp = 0.;
    init -> Vxp = 0.;
    init -> Vyp = 0.;
    init -> N = count = S->N;
  
#if 0
    if( T->b < .001 ) /* Orthographic */
      min_a += 0.;
    else
      min_a += T->zb / T->b;
#endif

    dp = S->depths;
    odp = init->depths;
    while( count-- ) {
      odp -> id = dp->id;
      odp -> a = ( dp->id == min_id ) ? min_a : ( dp->id == max_id ) ?
	min_a + 2. : min_a + 1.;
      odp -> Va = ( dp->id == min_id ) ? 0. : ( dp->id == max_id ) ? .05 : .1;
      dp++;
      odp++;
    }
    init -> print( init );
    init -> close( init );
  }

  first = 0;
}

void
  irand48()
{
  struct timeval the_time;
  gettimeofday( &the_time, NULL );
  srand48( the_time.tv_usec );
}

main(int argc, char **argv)
{
  fts_type	*Features;
  est_type	*Trajectory;
  sta_type	*Shape;
  FLOATTYPE	*Biases;
  int		first_frame = 1, frame = 0;

  ParseCmdLine(argc, argv);

  irand48();

  RotMat = make_matrix_type( 3, 3, "Rotation Matrix" );

  /* INPUTS */
  Shape = sta_open_read(params.sta_name);
  Trajectory = est_open_read(params.est_name);

  /* OUTPUT */
  Features = fts_open_write(params.fts_name);

  Shape->scan(Shape);
  Biases = compute_biases(Shape, params.bias_level);

  if( params.single ) {
    frame = 0;
    while( EOF != Trajectory->scan(Trajectory) && frame++ != params.frame )
      first_frame = 0;
    compute_features( Shape, Trajectory, Features, Biases, first_frame );
    Features->print( Features );
  }
  else
    while( EOF != Trajectory->scan(Trajectory) ) {
      compute_features(Shape, Trajectory, Features, Biases, first_frame);
      first_frame = 0;
      Features->print(Features);
    }
  exit(0);
}
